'''
Save data in ./data/pkl_data/xxx
'''

import pickle


# ---------- common
def load_input():
    with open('./data/pkl_data/input_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_input(rin):
    with open('./data/pkl_data/input_data.pkl', 'wb') as f:
        pickle.dump(rin, f)


def load_init_struc():
    with open('./data/pkl_data/init_struc_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_init_struc(init_struc_data):
    with open('./data/pkl_data/init_struc_data.pkl', 'wb') as f:
        pickle.dump(init_struc_data, f)


def load_opt_struc():
    with open('./data/pkl_data/opt_struc_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_opt_struc(opt_struc_data):
    with open('./data/pkl_data/opt_struc_data.pkl', 'wb') as f:
        pickle.dump(opt_struc_data, f)


def load_id_queueing():
    with open('./data/pkl_data/id_queueing.pkl', 'rb') as f:
        return pickle.load(f)


def save_id_queueing(id_queueing):
    with open('./data/pkl_data/id_queueing.pkl', 'wb') as f:
        pickle.dump(id_queueing, f)


def load_id_running():
    with open('./data/pkl_data/id_running.pkl', 'rb') as f:
        return pickle.load(f)


def save_id_running(id_running):
    with open('./data/pkl_data/id_running.pkl', 'wb') as f:
        pickle.dump(id_running, f)


def load_rslt():
    with open('./data/pkl_data/rslt_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_rslt(rslt_data):
    with open('./data/pkl_data/rslt_data.pkl', 'wb') as f:
        pickle.dump(rslt_data, f)


def load_kpt():
    with open('./data/pkl_data/kpt_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_kpt(kpt_data):
    with open('./data/pkl_data/kpt_data.pkl', 'wb') as f:
        pickle.dump(kpt_data, f)


def load_energy_step():
    with open('./data/pkl_data/energy_step_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_energy_step(energy_step_data):
    with open('./data/pkl_data/energy_step_data.pkl', 'wb') as f:
        pickle.dump(energy_step_data, f)


def load_struc_step():
    with open('./data/pkl_data/struc_step_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_struc_step(struc_step_data):
    with open('./data/pkl_data/struc_step_data.pkl', 'wb') as f:
        pickle.dump(struc_step_data, f)


def load_force_step():
    with open('./data/pkl_data/force_step_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_force_step(force_step_data):
    with open('./data/pkl_data/force_step_data.pkl', 'wb') as f:
        pickle.dump(force_step_data, f)


def load_stress_step():
    with open('./data/pkl_data/stress_step_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_stress_step(stress_step_data):
    with open('./data/pkl_data/stress_step_data.pkl', 'wb') as f:
        pickle.dump(stress_step_data, f)


# ---------- BO
def load_n_selection():
    with open('./data/pkl_data/n_selection.pkl', 'rb') as f:
        return pickle.load(f)


def save_n_selection(n_selection):
    with open('./data/pkl_data/n_selection.pkl', 'wb') as f:
        pickle.dump(n_selection, f)


def load_init_dscrpt_data():
    with open('./data/pkl_data/init_dscrpt_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_init_dscrpt_data(init_dscrpt_data):
    with open('./data/pkl_data/init_dscrpt_data.pkl', 'wb') as f:
        pickle.dump(init_dscrpt_data, f)


def load_opt_dscrpt_data():
    with open('./data/pkl_data/opt_dscrpt_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_opt_dscrpt_data(opt_dscrpt_data):
    with open('./data/pkl_data/opt_dscrpt_data.pkl', 'wb') as f:
        pickle.dump(opt_dscrpt_data, f)


def load_bo_mean():
    with open('./data/pkl_data/bo_mean.pkl', 'rb') as f:
        return pickle.load(f)


def save_bo_mean(bo_mean):
    with open('./data/pkl_data/bo_mean.pkl', 'wb') as f:
        pickle.dump(bo_mean, f)


def load_bo_var():
    with open('./data/pkl_data/bo_var.pkl', 'rb') as f:
        return pickle.load(f)


def save_bo_var(bo_var):
    with open('./data/pkl_data/bo_var.pkl', 'wb') as f:
        pickle.dump(bo_var, f)


def load_bo_score():
    with open('./data/pkl_data/bo_score.pkl', 'rb') as f:
        return pickle.load(f)


def save_bo_score(bo_score):
    with open('./data/pkl_data/bo_score.pkl', 'wb') as f:
        pickle.dump(bo_score, f)


# ---------- BO and LAQA
def load_id_select_hist():
    with open('./data/pkl_data/id_select_hist.pkl', 'rb') as f:
        return pickle.load(f)


def save_id_select_hist(id_select_hist):
    with open('./data/pkl_data/id_select_hist.pkl', 'wb') as f:
        pickle.dump(id_select_hist, f)


# ---------- LAQA
def load_tot_step_select():
    with open('./data/pkl_data/tot_step_select.pkl', 'rb') as f:
        return pickle.load(f)


def save_tot_step_select(tot_step_select):
    with open('./data/pkl_data/tot_step_select.pkl', 'wb') as f:
        pickle.dump(tot_step_select, f)


def load_laqa_step():
    with open('./data/pkl_data/laqa_step.pkl', 'rb') as f:
        return pickle.load(f)


def save_laqa_step(laqa_step):
    with open('./data/pkl_data/laqa_step.pkl', 'wb') as f:
        pickle.dump(laqa_step, f)


def load_laqa_struc():
    with open('./data/pkl_data/laqa_struc.pkl', 'rb') as f:
        return pickle.load(f)


def save_laqa_struc(laqa_struc):
    with open('./data/pkl_data/laqa_struc.pkl', 'wb') as f:
        pickle.dump(laqa_struc, f)


def load_laqa_energy():
    with open('./data/pkl_data/laqa_energy.pkl', 'rb') as f:
        return pickle.load(f)


def save_laqa_energy(laqa_energy):
    with open('./data/pkl_data/laqa_energy.pkl', 'wb') as f:
        pickle.dump(laqa_energy, f)


def load_laqa_bias():
    with open('./data/pkl_data/laqa_bias.pkl', 'rb') as f:
        return pickle.load(f)


def save_laqa_bias(laqa_bias):
    with open('./data/pkl_data/laqa_bias.pkl', 'wb') as f:
        pickle.dump(laqa_bias, f)


def load_laqa_score():
    with open('./data/pkl_data/laqa_score.pkl', 'rb') as f:
        return pickle.load(f)


def save_laqa_score(laqa_score):
    with open('./data/pkl_data/laqa_score.pkl', 'wb') as f:
        pickle.dump(laqa_score, f)


# ---------- EA
def load_gen():
    with open('./data/pkl_data/gen.pkl', 'rb') as f:
        return pickle.load(f)


def save_gen(gen):
    with open('./data/pkl_data/gen.pkl', 'wb') as f:
        pickle.dump(gen, f)


def load_struc_mol_id():
    with open('./data/pkl_data/struc_mol_id_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_struc_mol_id(struc_mol_id):
    with open('./data/pkl_data/struc_mol_id_data.pkl', 'wb') as f:
        pickle.dump(struc_mol_id, f)


def load_elite_struc():
    with open('./data/pkl_data/elite_struc.pkl', 'rb') as f:
        return pickle.load(f)


def save_elite_struc(elite_struc):
    with open('./data/pkl_data/elite_struc.pkl', 'wb') as f:
        pickle.dump(elite_struc, f)


def load_elite_fitness():
    with open('./data/pkl_data/elite_fitness.pkl', 'rb') as f:
        return pickle.load(f)


def save_elite_fitness(elite_fitness):
    with open('./data/pkl_data/elite_fitness.pkl', 'wb') as f:
        pickle.dump(elite_fitness, f)


def load_ea_info():
    with open('./data/pkl_data/ea_info.pkl', 'rb') as f:
        return pickle.load(f)


def save_ea_info(ea_info):
    with open('./data/pkl_data/ea_info.pkl', 'wb') as f:
        pickle.dump(ea_info, f)


def load_ea_origin():
    with open('./data/pkl_data/ea_origin.pkl', 'rb') as f:
        return pickle.load(f)


def save_ea_origin(ea_origin):
    with open('./data/pkl_data/ea_origin.pkl', 'wb') as f:
        pickle.dump(ea_origin, f)


def load_nat_data():
    with open('./data/pkl_data/nat_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_nat_data(nat_data):
    with open('./data/pkl_data/nat_data.pkl', 'wb') as f:
        pickle.dump(nat_data, f)


def load_ratio_data():
    with open('./data/pkl_data/ratio_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_ratio_data(ratio_data):
    with open('./data/pkl_data/ratio_data.pkl', 'wb') as f:
        pickle.dump(ratio_data, f)


def load_hdist_data():
    with open('./data/pkl_data/hdist_data.pkl', 'rb') as f:
        return pickle.load(f)


def save_hdist_data(hdist_data):
    with open('./data/pkl_data/hdist_data.pkl', 'wb') as f:
        pickle.dump(hdist_data, f)


# ---------- ext
def load_ext_opt_struc():
    with open('./ext/calc_data/ext_opt_struc_data.pkl', 'rb') as f:
        return pickle.load(f)


def load_ext_energy():
    with open('./ext/calc_data/ext_energy_data.pkl', 'rb') as f:
        return pickle.load(f)



import numpy as np


class SelectParents:
    '''
    select parents

    # ---------- args
    ranking (list): [ID, ...], ranking of IDs based on fitness without duplication

    # ---------- instance methods
    self.set_tournament(self, t_size)
    self.set_roulette(self, fitness, a_rlt, b_rlt, fit_reverse)
    self.get_parents(n_parent) <-- self._tournament or self._roulette
    '''

    def __init__(self, ranking):
        # ---------- self
        self.ranking = ranking

    def set_tournament(self, t_size):
        '''
        setting for tournament selection
        '''
        # ---------- self
        self.t_size = t_size
        # ---------- set self.get_parents() <-- self._tournament or self._roulette
        self.get_parents = self._tournament

    def _tournament(self, n_parent):
        '''
        tournament selection

        # ---------- args
        n_parent (int): number of parents

        # ---------- return
        parent_id (list): ID of selected parents
        '''
        parent_id = []
        while len(parent_id) < n_parent:
            t_indx = np.random.choice(len(self.ranking), self.t_size, replace=False)
            if parent_id:    # not allow the same parent in crossover
                if parent_id[0] == self.ranking[min(t_indx)]:
                    continue
            parent_id.append(self.ranking[min(t_indx)])
        return parent_id

    def set_roulette(self, fitness, a_rlt, b_rlt, fit_reverse):
        '''
        setting for roulette selection

        # ---------- args
        fitness (dict) : {ID: fitness, ...}
        a_rlt (float): a parameter in linear scaling
        b_rlt (float): b parameter in linear scaling
        fit_reverse (bool): if False, lower fitness is better

        # ---------- instance variables
        self.cum_fit (np.array): cumulative fitness
        '''
        # ----------self
        self.a_rlt = a_rlt
        self.b_rlt = b_rlt
        self.fit_reverse = fit_reverse
        # ---------- calculate cumulative fitness
        fit_deduped = np.array([fitness[i] for i in self.ranking])
        fit_deduped = self._linear_scaling(fit_deduped)
        self.cum_fit = np.cumsum(fit_deduped/fit_deduped.sum())
        # ---------- set self.get_parents <-- self._tournament or self._roulette
        self.get_parents = self._roulette

    def _roulette(self, n_parent):
        '''
        roulette selection

        # ---------- args
        n_parent (int): number of parents

        # ---------- return
        parent_id (list): ID of selected parents
        '''
        # ---------- select parents
        parent_id = []
        while len(parent_id) < n_parent:
            # indx_array: if all false, array is vacant
            indx_array = np.where(self.cum_fit < np.random.rand())[0]
            # select_indx: consider vacant or not
            select_indx = indx_array[-1] + 1 if indx_array.size != 0 else 0
            if parent_id:    # not allow same parent for crossover
                if parent_id[0] == self.ranking[select_indx]:
                    continue
            parent_id.append(self.ranking[select_indx])
        return parent_id

    def _linear_scaling(self, fit_deduped):
        # ------ in case of int
        a = float(self.a_rlt)
        b = float(self.b_rlt)
        # ---------- for fit_reverse
        if not self.fit_reverse:
            fit_deduped = -fit_deduped
        # ---------- scaling
        fmax = float(fit_deduped.max())
        fmin = float(fit_deduped.min())
        # ------ in case the same values
        if fmax == fmin:
            return fit_deduped
        fit_deduped = (a - b)/(fmax - fmin)*fit_deduped + (
            (b*fmax - a*fmin)/(fmax - fmin))
        # ---------- return
        return fit_deduped



'''
Control jobs in Quantum ESPRESSO
'''

from logging import getLogger
import os
import shutil

from pymatgen.io.vasp import Kpoints

from . import structure as qe_structure
from ...IO.out_results import out_kpts
from ...IO import pkl_data


logger = getLogger('cryspy')

def next_stage_qe(rin, stage, work_path, nat, kpt_data, cid):
    # ---------- skip_flag
    skip_flag = False

    # ---------- rename QE files at the current stage
    qe_files = [rin.qe_infile, rin.qe_outfile]
    for file in qe_files:
        if not os.path.isfile(work_path + file):
            logger.error('Not found ' + work_path + file)
            raise SystemExit(1)
        os.rename(work_path + file, work_path + f'stage{stage}_' + file)

    # ---------- next structure
    try:
        lines_cell = qe_structure.extract_cell_parameters(
            work_path+f'stage{stage}_'+rin.qe_outfile)
        if lines_cell is None:
            lines_cell = qe_structure.extract_cell_parameters(
                work_path+f'stage{stage}_'+rin.qe_infile)
        lines_atom = qe_structure.extract_atomic_positions(
            work_path+f'stage{stage}_'+rin.qe_outfile, nat)
        if lines_atom is None:
            lines_atom = qe_structure.extract_atomic_positions(
                work_path+f'stage{stage}_'+rin.qe_infile, nat)
        structure = qe_structure.from_lines(lines_cell, lines_atom)
    except ValueError:
        skip_flag = True
        kpt_data[cid].append(['skip'])
        pkl_data.save_kpt(kpt_data)
        out_kpts(kpt_data)
        logger.info(f'    error in QE,  skip structure {cid}')
        return skip_flag, kpt_data

    # ---------- copy the input file from ./calc_in for the next stage
    finput = './calc_in/' + rin.qe_infile + f'_{stage + 1}'
    shutil.copyfile(finput, work_path+rin.qe_infile)

    # ---------- "nat" (<-- natot) in qe_infile for EA-vc
    if rin.algo == 'EA-vc':
        _replace_nat(work_path + rin.qe_infile, nat)

    # ---------- append structure info.
    with open(work_path+rin.qe_infile, 'a') as fin:
        fin.write('\n')
    qe_structure.write(structure, work_path+rin.qe_infile, mode='a')

    # ---------- K_POINTS
    # kppvol[0]: <--> stage 1, kppvol[1] <--> stage2, ...
    #   so (stage - 1): current stage, stage: next stage in kppvol
    kpoints = Kpoints.automatic_density_by_vol(structure=structure,
                                               kppvol=rin.kppvol[stage])
    with open(work_path+rin.qe_infile, 'a') as f:
        f.write('\n')
        f.write('K_POINTS automatic\n')
        f.write(' '.join(str(x) for x in kpoints.kpts[0]) + '  0 0 0\n')

    # ---------- kpt_data
    kpt_data[cid].append(kpoints.kpts[0])
    pkl_data.save_kpt(kpt_data)
    out_kpts(kpt_data)

    # ---------- return
    return skip_flag, kpt_data


def next_struc_qe(rin, structure, cid, work_path, nat, kpt_data):
    # ---------- copy files
    calc_inputs = [rin.qe_infile]
    for f in calc_inputs:
        ff = f+'_1' if f == rin.qe_infile else f
        if not os.path.isfile('./calc_in/' + ff):
            logger.error('Could not find ./calc_in/' + ff)
            raise SystemExit(1)
        shutil.copyfile('./calc_in/'+ff, work_path+f)

    # ---------- "nat" (<-- natot) in qe_infile for EA-vc
    if rin.algo == 'EA-vc':
        _replace_nat(work_path + rin.qe_infile, nat)

    # ---------- append structure info. to the input file
    with open(work_path+rin.qe_infile, 'a') as fin:
        fin.write('\n')
    qe_structure.write(structure, work_path+rin.qe_infile, mode='a')

    # ---------- K_POINTS
    kpoints = Kpoints.automatic_density_by_vol(structure=structure,
                                               kppvol=rin.kppvol[0])
    with open(work_path+rin.qe_infile, 'a') as f:
        f.write('\n')
        f.write('K_POINTS automatic\n')
        f.write(' '.join(str(x) for x in kpoints.kpts[0]) + '  0 0 0\n')

    # ---------- kpt_data
    kpt_data[cid] = []    # initialize
    kpt_data[cid].append(kpoints.kpts[0])
    pkl_data.save_kpt(kpt_data)
    out_kpts(kpt_data)

    # ---------- return
    return kpt_data


def _replace_nat(filename, nat):
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines2 = []
    for line in lines:
        if line.lstrip().startswith('nat'):
            lines2.append(f'    nat = {sum(nat)}\n')
        else:
            lines2.append(line)
    with open(filename, 'w') as f:
        f.writelines(lines2)


from logging import getLogger

# ---------- import later
# from .VASP import calc_files_vasp, ctrl_job_vasp, collect_vasp
# from .QE import calc_files_qe, ctrl_job_qe, collect_qe
# from .OMX import calc_files_OMX, ctrl_job_OMX, collect_OMX
# from .soiap import calc_files_soiap, ctrl_job_soiap, collect_soiap
# from .LAMMPS import calc_files_lammps, ctrl_job_lammps, collect_lammps
# from .ASE import calc_files_ase, ctrl_job_ase, collect_ase
# from .ext import collect_ext


logger = getLogger('cryspy')


def check_calc_files(rin):
    if rin.calc_code == 'VASP':
        from .VASP import calc_files_vasp
        calc_files_vasp.check_input_vasp(rin)
    elif rin.calc_code == 'QE':
        from .QE import calc_files_qe
        calc_files_qe.check_input_qe(rin)
    elif rin.calc_code == 'OMX':
        from .OMX import calc_files_OMX
        calc_files_OMX.check_input_OMX(rin)
    elif rin.calc_code == 'soiap':
        from .soiap import calc_files_soiap
        calc_files_soiap.check_input_soiap(rin)
    elif rin.calc_code == 'LAMMPS':
        from .LAMMPS import calc_files_lammps
        calc_files_lammps.check_input_lammps(rin)
    elif rin.calc_code == 'ASE':
        from .ASE import calc_files_ase
        calc_files_ase.check_input_ase(rin)
    elif rin.calc_code == 'ext':
        pass
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def next_stage(rin, stage, work_path, nat, kpt_data=None, cid=None):
    if rin.calc_code == 'VASP':
        from .VASP import ctrl_job_vasp
        skip_flag, kpt_data = ctrl_job_vasp.next_stage_vasp(rin, stage, work_path,
                                                            kpt_data, cid)
        return skip_flag, kpt_data
    elif rin.calc_code == 'QE':
        from .QE import ctrl_job_qe
        # qe needs the number of atoms
        skip_flag, kpt_data = ctrl_job_qe.next_stage_qe(rin, stage, work_path,
                                                        nat, kpt_data, cid)
        return skip_flag, kpt_data
    elif rin.calc_code == 'OMX':
        from .OMX import ctrl_job_OMX
        # omx needs the number of atoms
        skip_flag, kpt_data = ctrl_job_OMX.next_stage_OMX(rin, stage, work_path,
                                                          nat, kpt_data, cid)
        return skip_flag, kpt_data
    elif rin.calc_code == 'soiap':
        from .soiap import ctrl_job_soiap
        # soiap needs the number of atoms
        skip_flag = ctrl_job_soiap.next_stage_soiap(rin, stage, work_path, nat)
        return skip_flag
    elif rin.calc_code == 'LAMMPS':
        from .LAMMPS import ctrl_job_lammps
        # lammps needs the number of atoms
        skip_flag = ctrl_job_lammps.next_stage_lammps(rin, stage, work_path, nat)
        return skip_flag
    elif rin.calc_code == 'ASE':
        from .ASE import ctrl_job_ase
        skip_flag = ctrl_job_ase.next_stage_ase(rin, stage, work_path)
        return skip_flag
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def next_struc(rin, structure, cid, work_path, nat, kpt_data=None):
    # args[0] <-- kpt_data
    if rin.calc_code == 'VASP':
        from .VASP import ctrl_job_vasp
        kpt_data = ctrl_job_vasp.next_struc_vasp(rin, structure, cid,
                                                 work_path, kpt_data)
        return kpt_data
    elif rin.calc_code == 'QE':
        from .QE import ctrl_job_qe
        # qe needs the number of atoms
        kpt_data = ctrl_job_qe.next_struc_qe(rin, structure, cid,
                                             work_path, nat, kpt_data)
        return kpt_data
    elif rin.calc_code == 'OMX':
        from .OMX import ctrl_job_OMX
        kpt_data = ctrl_job_OMX.next_struc_OMX(rin, structure, cid,
                                               work_path, nat, kpt_data)
        return kpt_data
    elif rin.calc_code == 'soiap':
        from .soiap import ctrl_job_soiap
        ctrl_job_soiap.next_struc_soiap(rin, structure, cid, work_path)
    elif rin.calc_code == 'LAMMPS':
        from .LAMMPS import ctrl_job_lammps
        # lammps needs the number of atoms
        ctrl_job_lammps.next_struc_lammps(rin, structure, cid, work_path ,nat)
    elif rin.calc_code == 'ASE':
        from .ASE import ctrl_job_ase
        ctrl_job_ase.next_struc_ase(rin, structure, cid, work_path)
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def collect(rin, cid, work_path, nat):
    if rin.calc_code == 'VASP':
        from .VASP import collect_vasp
        # vasp needs the number of atoms to collect results
        # rin is not used in collect_vasp
        opt_struc, energy, magmom, check_opt = \
            collect_vasp.collect_vasp(cid, work_path, nat)
    elif rin.calc_code == 'QE':
        from .QE import collect_qe
        # qe needs the number of atoms to collect results
        opt_struc, energy, magmom, check_opt = \
            collect_qe.collect_qe(rin, cid, work_path, nat)
    elif rin.calc_code == 'OMX':
        from .OMX import collect_OMX
        # OMX needs the number of atoms to collect results
        opt_struc, energy, magmom, check_opt = \
            collect_OMX.collect_OMX(rin, cid, work_path, nat)
    elif rin.calc_code == 'soiap':
        from .soiap import collect_soiap
        # soiap needs the number of atoms to collect results
        opt_struc, energy, magmom, check_opt = \
            collect_soiap.collect_soiap(rin, cid, work_path, nat)
    elif rin.calc_code == 'LAMMPS':
        from .LAMMPS import collect_lammps
        # lammps needs the number of atoms to collect results
        opt_struc, energy, magmom, check_opt = \
            collect_lammps.collect_lammps(rin, cid, work_path, nat)
    elif rin.calc_code == 'ASE':
        from .ASE import collect_ase
        # ase needs the number of atoms to collect results
        # rin is not used in collect_ase
        opt_struc, energy, magmom, check_opt = \
            collect_ase.collect_ase(cid, work_path, nat)
    elif rin.calc_code == 'ext':
        from .ext import collect_ext
        ext_opt_struc_data, ext_energy_data = collect_ext.collect_ext()
        return ext_opt_struc_data, ext_energy_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)

    # ---------- return
    return opt_struc, energy, magmom, check_opt


def get_energy_step(rin, energy_step_data, cid, work_path, nat):
    from .VASP import collect_vasp
    if rin.calc_code == 'VASP':
        # vasp needs the number of atoms to collect results
        # rin is not used in collect_vasp
        energy_step_data = collect_vasp.get_energy_step_vasp(
            energy_step_data, cid, work_path, nat)
        return energy_step_data
    elif rin.calc_code == 'QE':
        from .QE import collect_qe
        # qe needs the number of atoms to collect results
        energy_step_data = collect_qe.get_energy_step_qe(rin, energy_step_data,
                                                         cid, work_path, nat)
        return energy_step_data
    elif rin.calc_code == 'soiap':
        from .soiap import collect_soiap
        # rin is not used in collect_soiap.get_energy_step_soiap
        energy_step_data = collect_soiap.get_energy_step_soiap(
            energy_step_data, cid, work_path)
        return energy_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_struc_step(rin, struc_step_data, cid, work_path, nat):
    if rin.calc_code == 'VASP':
        from .VASP import collect_vasp
        # rin is not used in collect_vasp
        struc_step_data = collect_vasp.get_struc_step_vasp(
            struc_step_data, cid, work_path)
        return struc_step_data
    elif rin.calc_code == 'QE':
        from .QE import collect_qe
        # qe needs the number of atoms to collect results
        struc_step_data = collect_qe.get_struc_step_qe(
            rin, struc_step_data, cid, work_path, nat)
        return struc_step_data
    elif rin.calc_code == 'soiap':
        from .soiap import collect_soiap
        # soiap needs the number of atoms to collect results
        struc_step_data = collect_soiap.get_struc_step_soiap(
            rin, struc_step_data, cid, work_path, nat)
        return struc_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_force_step(rin, force_step_data, cid, work_path, nat):
    if rin.calc_code == 'VASP':
        from .VASP import collect_vasp
        # rin is not used in collect_vasp
        force_step_data = collect_vasp.get_force_step_vasp(
            force_step_data, cid, work_path)
        return force_step_data
    elif rin.calc_code == 'QE':
        from .QE import collect_qe
        # qe needs the number of atoms to collect results
        force_step_data = collect_qe.get_force_step_qe(
            rin, force_step_data, cid, work_path, nat)
        return force_step_data
    elif rin.calc_code == 'soiap':
        from .soiap import collect_soiap
        # soiap needs the number of atoms to collect results
        # rin is not used in collect_soiap.get_force_step_soiap
        force_step_data = collect_soiap.get_force_step_soiap(
            force_step_data, cid, work_path, nat)
        return force_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_stress_step(rin, stress_step_data, cid, work_path):
    if rin.calc_code == 'VASP':
        from .VASP import collect_vasp
        # rin is not used in collect_vasp
        stress_step_data = collect_vasp.get_stress_step_vasp(
            stress_step_data, cid, work_path)
        return stress_step_data
    elif rin.calc_code == 'QE':
        from .QE import collect_qe
        stress_step_data = collect_qe.get_stress_step_qe(
            rin, stress_step_data, cid, work_path)
        return stress_step_data
    elif rin.calc_code == 'soiap':
        from .soiap import collect_soiap
        # rin is not used in collect_soiap.get_stress_step_soiap
        stress_step_data = collect_soiap.get_stress_step_soiap(
            stress_step_data, cid, work_path)
        return stress_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)



import itertools
from logging import getLogger
import os
import shutil
import subprocess

import numpy as np

from ..interface import select_code
from ..IO import change_input, io_stat, pkl_data
from ..IO.out_results import out_rslt, out_hdist
from ..util.utility import backup_cryspy
from ..util.struc_util import out_poscar, out_cif

# ---------- import later
#from ..BO.select_descriptor import select_descriptor
#from ..BO import bo_next_select
#from ..EA import ea_next_gen
#from ..EA.calc_ef import calc_ef
#from ..EA.calc_hull import calc_convex_hull_2d
#from ..LAQA.calc_score import calc_laqa_bias
#from ..LAQA import laqa_next_selection
#from ..IO.out_results import out_laqa_status, out_laqa_step, out_laqa_score
#from ..IO.out_results import out_laqa_energy, out_laqa_bias


logger = getLogger('cryspy')


class Ctrl_job:

    def __init__(self, rin, init_struc_data):
        self.rin = rin
        self.init_struc_data = init_struc_data
        self.opt_struc_data = pkl_data.load_opt_struc()
        self.rslt_data = pkl_data.load_rslt()
        self.id_queueing = pkl_data.load_id_queueing()
        self.id_running = pkl_data.load_id_running()
        self.recheck = False
        self.logic_next = False
        # ---------- for each algorithm
        if self.rin.algo == 'RS':
            pass
        elif self.rin.algo == 'BO':
            self.n_selection = pkl_data.load_n_selection()
            self.id_select_hist = pkl_data.load_id_select_hist()
            self.init_dscrpt_data = pkl_data.load_init_dscrpt_data()
            self.opt_dscrpt_data = pkl_data.load_opt_dscrpt_data()
            self.bo_mean = pkl_data.load_bo_mean()
            self.bo_var = pkl_data.load_bo_var()
            self.bo_score = pkl_data.load_bo_score()
        elif self.rin.algo == 'LAQA':
            self.id_select_hist = pkl_data.load_id_select_hist()
            self.tot_step_select = pkl_data.load_tot_step_select()
            self.laqa_step = pkl_data.load_laqa_step()
            self.laqa_struc = pkl_data.load_laqa_struc()
            self.laqa_energy = pkl_data.load_laqa_energy()
            self.laqa_bias = pkl_data.load_laqa_bias()
            self.laqa_score = pkl_data.load_laqa_score()
        elif self.rin.algo in ['EA', 'EA-vc']:
            self.gen = pkl_data.load_gen()
            if self.rin.struc_mode in ['mol', 'mol_bs']:
                self.struc_mol_id = pkl_data.load_struc_mol_id()
            if self.rin.algo == 'EA-vc':
                self.nat_data = pkl_data.load_nat_data()
                self.ratio_data = pkl_data.load_ratio_data()
                self.hdist_data = pkl_data.load_hdist_data()
            # do not have to load ea_data here.
            # ea_data is used only in ea_next_gen.py
        # ---------- for option
        if self.rin.kpt_flag:
            self.kpt_data = pkl_data.load_kpt()
        if self.rin.energy_step_flag:
            self.energy_step_data = pkl_data.load_energy_step()
        if self.rin.struc_step_flag:
            self.struc_step_data = pkl_data.load_struc_step()
        if self.rin.force_step_flag:
            self.force_step_data = pkl_data.load_force_step()
        if self.rin.stress_step_flag:
            self.stress_step_data = pkl_data.load_stress_step()
        # ---------- flag for next selection or generation
        if self.rin.algo in ['BO', 'LAQA', 'EA', 'EA-vc']:
            if (self.id_queueing or self.id_running):
                self.go_next_sg = False
            else:
                self.go_next_sg = True

    def check_job(self):
        # ---------- option: recalc
        if self.rin.recalc is not None:
            self.set_recalc()
        # ---------- temporarily append
        self.tmp_running = self.id_running[:]    # shallow copy
        self.tmp_queueing = self.id_queueing[:]
        if not self.rin.stop_next_struc:    # if true --> option: stop_next_struc
            while len(self.tmp_running) < self.rin.njob and self.tmp_queueing:
                self.tmp_running.append(self.tmp_queueing.pop(0))
        # ---------- initialize
        self.stage_stat = {}    # key: Structure ID
        self.job_stat = {}
        # ---------- check job status
        for cid in self.tmp_running[:self.rin.njob]:
            # ------ mkdir
            if not os.path.isdir(f'work/{cid}'):
                os.mkdir(f'work/{cid}')
            # ------ check stat_job file
            stat_path = f'work/{cid}' + '/stat_job'
            try:
                with open(stat_path, 'r') as fstat:
                    istat = fstat.readline()    # id
                    sstat = fstat.readline()    # stage
                    jstat = fstat.readline()    # submitted or done or ...
                self.stage_stat[cid] = int(sstat.split()[0])
                if not cid == int(istat.split()[0]):
                    logger.error(f'ID is wrong in work/{cid}')
                    raise SystemExit(1)
                self.stage_stat[cid] = int(sstat.split()[0])
                if jstat[0:3] == 'sub':
                    self.job_stat[cid] = 'submitted'
                elif jstat[0:4] == 'done':
                    self.job_stat[cid] = 'done'
                elif jstat[0:4] == 'skip':
                    self.job_stat[cid] = 'skip'
                else:
                    self.job_stat[cid] = 'else'
            except IOError:
                self.stage_stat[cid] = 'no_file'
                self.job_stat[cid] = 'no_file'

    def set_recalc(self):
        # ---------- check id
        for tid in self.rin.recalc:
            if tid not in self.opt_struc_data:
                logger.error(f'ID {tid} has not yet been calculated')
                raise SystemExit(1)
        # ---------- append IDs to the head of id_queueing
        self.id_queueing = list(self.rin.recalc) + self.id_queueing
        # ---------- save and update id_queueing
        pkl_data.save_id_queueing(self.id_queueing)
        stat = io_stat.stat_read()
        io_stat.set_id(stat, 'id_queueing', self.id_queueing)
        io_stat.write_stat(stat)
        # ---------- log and out
        logger.info('# -- Recalc')
        logger.info(f'Append {self.rin.recalc} to the head of id_queueing')
        # ---------- clear recalc
        self.rin.recalc = None
        config = change_input.read_config()
        change_input.change_input(config, 'option', 'recalc', '')    # clear
        change_input.write_config(config)
        logger.info('Clear recalc in cryspy.in')

    def handle_job(self):
        logger.info('# ---------- job status')
        for cid in self.tmp_running[:self.rin.njob]:
            # ---------- set work_path and cid
            self.work_path = f'./work/{cid}/'
            self.cid = cid
            # ---------- handle job
            if self.job_stat[cid] == 'submitted':
                logger.info(f'ID {cid:>6}: still queueing or running')
            elif self.job_stat[cid] == 'done':
                self.ctrl_done()
            elif self.job_stat[cid] == 'skip':
                self.ctrl_skip()
            elif self.job_stat[cid] == 'else':
                logger.error(f'Wrong job_stat in {self.work_path}. ')
            elif self.job_stat[cid] == 'no_file':
                self.ctrl_next_struc()
            else:
                logger.error(f'Unexpected error in {self.work_path}stat_job')

    def ctrl_done(self):
        self.cstage = self.stage_stat[self.cid]
        # ---------- log
        logger.info(f'ID {self.cid:>6}: Stage {self.cstage} Done!')
        # ---------- next stage
        if self.cstage < self.rin.nstage:
            self.ctrl_next_stage()
        # ---------- collect result
        elif self.cstage == self.rin.nstage:
            self.ctrl_collect()
        # ---------- error
        else:
            logger.error('Wrong stage in '+self.work_path+'stat_job')
            raise SystemExit(1)

    def ctrl_next_stage(self):
        # ---------- EA-vc
        if self.rin.algo == 'EA-vc':
            nat = self.nat_data[self.cid]
        else:
            nat = self.rin.nat
        # ---------- energy step
        if self.rin.energy_step_flag:
            self.energy_step_data = select_code.get_energy_step(
                self.rin, self.energy_step_data, self.cid, self.work_path, nat)
        # ---------- struc step
        if self.rin.struc_step_flag:
            self.struc_step_data = select_code.get_struc_step(
                self.rin, self.struc_step_data, self.cid, self.work_path, nat)
        # ---------- force step
        if self.rin.force_step_flag:
            self.force_step_data = select_code.get_force_step(
                self.rin, self.force_step_data, self.cid, self.work_path, nat)
        # ---------- stress step
        if self.rin.stress_step_flag:
            self.stress_step_data = select_code.get_stress_step(
                self.rin, self.stress_step_data, self.cid, self.work_path)
        # ---------- next stage
        if self.rin.kpt_flag:
            skip_flag, self.kpt_data = select_code.next_stage(
                self.rin, self.cstage, self.work_path, nat,
                kpt_data=self.kpt_data, cid=self.cid)
        else:
            skip_flag = select_code.next_stage(self.rin, self.cstage,
                                               self.work_path, nat)
        # ---------- skip
        if skip_flag:
            self.ctrl_skip()
            return
        # ---------- prepare jobfile
        prepare_jobfile(self.rin, self.cid, self.work_path)
        # ---------- submit
        self.submit_next_stage()

    def submit_next_stage(self):
        # ---------- submit job
        os.chdir(self.work_path)    # cd work_path
        with open('stat_job', 'w') as fwstat:
            fwstat.write(f'{self.cid:<6}    # Structure ID\n')
            fwstat.write(f'{self.cstage + 1:<6}    # Stage\n')
            fwstat.write('submitted\n')
        with open('sublog', 'w') as logf:
            subprocess.Popen([self.rin.jobcmd, self.rin.jobfile],
                             stdout=logf, stderr=logf)
        os.chdir('../../')    # go back to ..
        # ---------- save status
        stat = io_stat.stat_read()
        io_stat.set_stage(stat, self.cid, self.cstage + 1)
        io_stat.write_stat(stat)
        # ---------- log
        logger.info(f'    submitted job, ID {self.cid:>6} Stage {self.cstage + 1}')

    def ctrl_collect(self):
        # ---------- EA-vc
        if self.rin.algo == 'EA-vc':
            nat = self.nat_data[self.cid]
        else:
            nat = self.rin.nat
        # ---------- energy step
        if self.rin.energy_step_flag:
            self.energy_step_data = select_code.get_energy_step(
                self.rin, self.energy_step_data, self.cid, self.work_path, nat)
        # ---------- struc step
        if self.rin.struc_step_flag:
            self.struc_step_data = select_code.get_struc_step(
                self.rin, self.struc_step_data, self.cid, self.work_path, nat)
        # ---------- force step
        if self.rin.force_step_flag:
            self.force_step_data = select_code.get_force_step(
                self.rin, self.force_step_data, self.cid, self.work_path, nat)
        # ---------- stress step
        if self.rin.stress_step_flag:
            self.stress_step_data = select_code.get_stress_step(
                self.rin, self.stress_step_data, self.cid, self.work_path)
        # ---------- each algo
        if self.rin.algo == 'RS':
            self.ctrl_collect_rs(nat)
        elif self.rin.algo == 'BO':
            self.ctrl_collect_bo(nat)
        elif self.rin.algo == 'LAQA':
            self.ctrl_collect_laqa(nat)
        elif self.rin.algo in ['EA', 'EA-vc']:
            self.ctrl_collect_ea(nat)
        else:
            logger.error('Error, algo')
            raise SystemExit(1)
        # ---------- move to fin
        if self.rin.algo == 'LAQA':
            if self.fin_laqa:
                mv_fin(self.cid)
            else:
                os.rename(self.work_path+'stat_job',
                          self.work_path+'prev_stat_job')
        else:
            mv_fin(self.cid)
        # ---------- update status
        self.id_queueing, self.id_running = update_status(
                                                self.cid,
                                                self.id_queueing,
                                                self.id_running,
                                                operation='fin',
                                            )
        # ---------- recheck
        self.recheck = True

    def ctrl_collect_rs(self, nat):
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.rin, self.cid, self.work_path, nat)
        logger.info(f'    collect results: E = {energy} eV/atom')
        # ---------- register opt data
        self.opt_struc_data, self.rslt_data = regist_opt(
            self.rin, self.cid, self.work_path,
            self.init_struc_data, self.opt_struc_data, self.rslt_data,
            opt_struc, energy, magmom, check_opt, ef=None, n_selection=None, gen=None)

    def ctrl_collect_bo(self, nat):
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.rin, self.cid, self.work_path, nat)
        logger.info(f'    collect results: E = {energy} eV/atom')
        # ---------- register opt data
        self.opt_struc_data, self.rslt_data = regist_opt(
            self.rin, self.cid, self.work_path,
            self.init_struc_data, self.opt_struc_data, self.rslt_data,
            opt_struc, energy, magmom, check_opt,
            ef=None, n_selection=self.n_selection, gen=None)
        # ---------- success
        if opt_struc is not None:
            from ..BO.select_descriptor import select_descriptor
            # ------ calc descriptor for opt sturcture
            tmp_dscrpt = select_descriptor(self.rin, {self.cid: opt_struc})
            # ------ update descriptors
            self.opt_dscrpt_data.update(tmp_dscrpt)
        # ---------- error
        else:
            # ------ update descriptors and non_error_id
            self.opt_dscrpt_data[self.cid] = None
        # ---------- save bo_data
        pkl_data.save_opt_dscrpt_data(self.opt_dscrpt_data)

    def ctrl_collect_laqa(self, nat):
        # ---------- flag for finish
        self.fin_laqa = False
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.rin, self.cid, self.work_path, nat)
        # ---------- total step and laqa_step
        #     force_step_data[ID][stage][step][atom]
        if self.force_step_data[self.cid][-1] is None:
            self.laqa_step[self.cid].append(0)
        else:
            self.tot_step_select[-1] += len(
                self.force_step_data[self.cid][-1])
            self.laqa_step[self.cid].append(
                len(self.force_step_data[self.cid][-1]))
        # ------ save status
        stat = io_stat.stat_read()
        io_stat.set_common(stat, 'total_step', sum(self.tot_step_select))
        io_stat.write_stat(stat)
        # ---------- append laqa struc
        self.laqa_struc[self.cid].append(opt_struc)
        # ---------- append laqa energy
        self.laqa_energy[self.cid].append(energy)
        # ---------- append laqa bias
        #     force_step_data[ID][stage][step][atom]
        #     stress_step_data[ID][stage][step][atom]
        from ..LAQA.calc_score import calc_laqa_bias
        tmp_laqa_bias = calc_laqa_bias(
            self.force_step_data[self.cid][-1],
            self.stress_step_data[self.cid][-1],
            wf=self.rin.wf, ws=self.rin.ws)
        self.laqa_bias[self.cid].append(tmp_laqa_bias)
        # ---------- append laqa score
        if check_opt == 'done' or np.isnan(energy) or np.isnan(tmp_laqa_bias):
            self.laqa_score[self.cid].append(-float('inf'))
        else:
            self.laqa_score[self.cid].append(-energy + tmp_laqa_bias)
        # ---------- save and out laqa data
        pkl_data.save_tot_step_select(self.tot_step_select)
        pkl_data.save_laqa_step(self.laqa_step)
        pkl_data.save_laqa_struc(self.laqa_struc)
        pkl_data.save_laqa_energy(self.laqa_energy)
        pkl_data.save_laqa_bias(self.laqa_bias)
        pkl_data.save_laqa_score(self.laqa_score)
        from ..IO.out_results import out_laqa_status, out_laqa_step, out_laqa_score
        from ..IO.out_results import out_laqa_energy, out_laqa_bias
        out_laqa_status(self.laqa_step, self.laqa_score,
                        self.laqa_energy, self.laqa_bias)
        out_laqa_step(self.laqa_step)
        out_laqa_score(self.laqa_score)
        out_laqa_energy(self.laqa_energy)
        out_laqa_bias(self.laqa_bias)
        # ---------- case of 'done' or error
        if check_opt == 'done' or np.isnan(energy) or np.isnan(tmp_laqa_bias):
            self.fin_laqa = True
            logger.info(f'    collect results: E = {energy} eV/atom')
            # ------ register opt data
            self.opt_struc_data, self.rslt_data = regist_opt(
                self.rin, self.cid, self.work_path,
                self.init_struc_data, self.opt_struc_data, self.rslt_data,
                opt_struc, energy, magmom, check_opt, ef=None, n_selection=None, gen=None)

    def ctrl_collect_ea(self, nat):
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.rin, self.cid, self.work_path, nat)
        logger.info(f'    collect results: E = {energy} eV/atom')
        # ---------- calculate Ef
        if self.rin.algo == 'EA-vc':
            from ..EA.calc_ef import calc_ef
            ef = calc_ef(energy, self.ratio_data[self.cid], self.rin.end_point)
            logger.info(f'                     Ef = {ef} eV/atom')
        else:
            ef = None
        # ---------- register opt data
        self.opt_struc_data, self.rslt_data = regist_opt(
            self.rin, self.cid, self.work_path,
            self.init_struc_data, self.opt_struc_data, self.rslt_data,
            opt_struc, energy, magmom, check_opt, ef=ef, n_selection=None, gen=self.gen)

    def ctrl_next_struc(self):
        # ---------- RS, BO, EA, EA-vc
        if self.rin.algo in ['RS', 'BO', 'EA', 'EA-vc']:
            next_struc_data = self.init_struc_data[self.cid]
        # ---------- LAQA
        elif self.rin.algo == 'LAQA':
            if self.laqa_struc[self.cid]:    # vacant list?
                next_struc_data = self.laqa_struc[self.cid][-1]
            else:
                next_struc_data = self.init_struc_data[self.cid]
        # ---------- algo is wrong
        else:
            logger.error('Error, algo in ctrl_next_struc')
            raise SystemExit(1)
        # ---------- nat for EA-vc
        if self.rin.algo == 'EA-vc':
            nat = self.nat_data[self.cid]
        else:
            nat = self.rin.nat
        # ---------- common part
        # ------ in case there is no initial strucure data
        if next_struc_data is None:
            logger.info(f'ID {self.cid:>6}: initial structure is None')
            self.ctrl_skip()
        # ------ normal initial structure data
        else:
            # -- prepare input files for structure optimization
            if self.rin.kpt_flag:
                self.kpt_data = select_code.next_struc(
                                    self.rin,
                                    next_struc_data,
                                    self.cid,
                                    self.work_path,
                                    nat,
                                    self.kpt_data,
                                )
            else:
                select_code.next_struc(
                    self.rin,
                    next_struc_data,
                    self.cid,
                    self.work_path,
                    nat,
                )
            # -- prepare jobfile
            prepare_jobfile(self.rin, self.cid, self.work_path)
            # -- submit
            submit_next_struc(self.rin, self.cid, self.work_path)
            logger.info(f'ID {self.cid:>6}: submit job, Stage 1')
            # -- update status
            self.id_queueing, self.id_running = update_status(
                                                    self.cid,
                                                    self.id_queueing,
                                                    self.id_running,
                                                    operation='submit'
                                                )

    def ctrl_skip(self):
        # ---------- log
        logger.info(f'ID {self.cid:>6}: Skip')
        # ---------- get initial spg info
        if self.init_struc_data[self.cid] is None:
            spg_sym = None
            spg_num = 0
        else:
            try:
                spg_sym, spg_num = self.init_struc_data[
                    self.cid].get_space_group_info(symprec=self.rin.symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
        # ---------- 'skip' for rslt
        spg_num_opt = 0
        spg_sym_opt = None
        energy = np.nan
        magmom = np.nan
        check_opt = 'skip'
        # ---------- register opt_struc
        self.opt_struc_data[self.cid] = None
        pkl_data.save_opt_struc(self.opt_struc_data)
        # ---------- RS
        if self.rin.algo == 'RS':
            # ------ save rslt
            self.rslt_data.loc[self.cid] = [spg_num, spg_sym,
                                            spg_num_opt, spg_sym_opt,
                                            energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
        # ---------- BO
        elif self.rin.algo == 'BO':
            # ------ save rslt
            self.rslt_data.loc[self.cid] = [self.n_selection,
                                            spg_num, spg_sym,
                                            spg_num_opt, spg_sym_opt,
                                            energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
            # ------ update and save descriptors
            self.opt_dscrpt_data[self.cid] = None
            pkl_data.save_opt_dscrpt_data(self.opt_dscrpt_data)
        # ---------- LAQA
        elif self.rin.algo == 'LAQA':
            # ------ save rslt
            self.rslt_data.loc[self.cid] = [spg_num, spg_sym,
                                            spg_num_opt, spg_sym_opt,
                                            energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
            # ---------- laqa data
            self.laqa_step[self.cid].append(0)
            self.laqa_struc[self.cid].append(None)
            self.laqa_energy[self.cid].append(energy)
            self.laqa_bias[self.cid].append(np.nan)
            self.laqa_score[self.cid].append(-float('inf'))
            # ---------- save and out laqa data
            #pkl_data.save_tot_step_select(self.tot_step_select)    # not used here
            pkl_data.save_laqa_step(self.laqa_step)
            pkl_data.save_laqa_struc(self.laqa_struc)
            pkl_data.save_laqa_energy(self.laqa_energy)
            pkl_data.save_laqa_bias(self.laqa_bias)
            pkl_data.save_laqa_score(self.laqa_score)
            from ..IO.out_results import out_laqa_status, out_laqa_step, out_laqa_score
            from ..IO.out_results import out_laqa_energy, out_laqa_bias
            out_laqa_status(self.laqa_step, self.laqa_score,
                            self.laqa_energy, self.laqa_bias)
            out_laqa_step(self.laqa_step)
            out_laqa_score(self.laqa_score)
            out_laqa_energy(self.laqa_energy)
            out_laqa_bias(self.laqa_bias)
        # ---------- EA
        elif self.rin.algo == 'EA':
            self.rslt_data.loc[self.cid] = [self.gen,
                                            spg_num, spg_sym,
                                            spg_num_opt, spg_sym_opt,
                                            energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
        elif self.rin.algo == 'EA-vc':
            ef = np.nan
            self.rslt_data.loc[self.cid] = [self.gen,
                                            spg_num, spg_sym,
                                            spg_num_opt, spg_sym_opt,
                                            energy, ef, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
        # ---------- move to fin
        mv_fin(self.cid)
        # ---------- update status
        self.id_queueing, self.id_running = update_status(
                                                self.cid,
                                                self.id_queueing,
                                                self.id_running,
                                                operation='fin',
                                            )
        # ---------- recheck
        self.recheck = True

    def next_sg(self, noprint=False):
        '''
        next selection or generation
        '''
        if self.rin.algo == 'BO':
            self.next_select_BO(noprint)
        if self.rin.algo == 'LAQA':
            self.next_select_LAQA()
        if self.rin.algo in ['EA', 'EA-vc']:
            if self.rin.algo == 'EA-vc':
                ea_vc_data = (self.nat_data, self.ratio_data, self.hdist_data)
            else:
                ea_vc_data = None
            next_gen_EA(
                self.rin,
                self.id_queueing,
                self.id_running,
                self.gen,
                self.go_next_sg,
                self.init_struc_data,
                self.opt_struc_data,
                self.rslt_data,
                ea_vc_data=ea_vc_data,
                struc_mol_id=None,
            )

    def next_select_BO(self, noprint=False):
        # ---------- log
        logger.info(f'\nDone selection {self.n_selection}')
        # ---------- done all structures
        if len(self.rslt_data) == self.rin.tot_struc:
            logger.info('\nDone all structures!')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- flag for next selection or generation
        if not self.go_next_sg:
            logger.info('\nBO is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- check point 3
        if self.rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: BO is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- max_select_bo
        if 0 < self.rin.max_select_bo <= self.n_selection:
            logger.info(f'\nReached max_select_bo: {self.rin.max_select_bo}')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- BO
        backup_cryspy()
        bo_data = (self.init_dscrpt_data, self.opt_dscrpt_data,
                   self.bo_mean, self.bo_var, self.bo_score)
        bo_id_data = (self.n_selection, self.id_queueing,
                      self.id_running, self.id_select_hist)
        from ..BO import bo_next_select
        bo_next_select.next_select(self.rin, self.rslt_data,
                                   bo_id_data, bo_data, noprint)

    def next_select_LAQA(self):
        # ---------- flag for next selection or generation
        if not self.go_next_sg:
            logger.info('\nLAQA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- check point 3
        if self.rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: LAQA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- selection of LAQA
        backup_cryspy()
        laqa_id_data = (self.id_queueing, self.id_running,
                        self.id_select_hist)
        laqa_data = (self.tot_step_select, self.laqa_step, self.laqa_struc,
                     self.laqa_energy, self.laqa_bias, self.laqa_score)
        from ..LAQA import laqa_next_selection
        laqa_next_selection.next_selection(self.rin, laqa_id_data, laqa_data)


#
#  End Ctrl_job class
#


def prepare_jobfile(rin, cid, work_path):
    if not os.path.isfile('./calc_in/' + rin.jobfile):
        logger.error('Could not find ./calc_in' + rin.jobfile)
        raise SystemExit(1)
    with open('./calc_in/' + rin.jobfile, 'r') as f:
        lines = f.readlines()
    lines2 = []
    for line in lines:
        lines2.append(line.replace('CrySPY_ID', str(cid)))
    with open(work_path + rin.jobfile, 'w') as f:
        f.writelines(lines2)


def submit_next_struc(rin, cid, work_path, wait=False):
    # ---------- submit job
    os.chdir(work_path)    # cd work_path
    # ---------- submit
    with open('sublog', 'w') as logf:
        if not wait:
            subprocess.Popen([rin.jobcmd, rin.jobfile],
                             stdout=logf, stderr=logf)
        else:    # wait
            subprocess.run([rin.jobcmd, rin.jobfile],
                           stdout=logf, stderr=logf)
    # ---------- write stat_job
    with open('stat_job', 'w') as f:
        f.write(f'{cid:<6}    # Structure ID\n')
        f.write(f'{1:<6}    # Stage\n')
        f.write('submitted\n')
    os.chdir('../../')    # go back to csp root dir


def regist_opt(
        rin,
        cid,
        work_path,
        init_struc_data,
        opt_struc_data,
        rslt_data,
        opt_struc,
        energy,
        magmom,
        check_opt,
        ef=None,
        n_selection=None,
        gen=None
    ):
    '''
    Common part in ctrl_collect_*
    '''
    # ---------- get initial spg info
    try:
        spg_sym, spg_num = init_struc_data[
            cid].get_space_group_info(symprec=rin.symprec)
    except TypeError:
        spg_num = 0
        spg_sym = None

    # ---------- success
    if opt_struc is not None:
        # ------ get opt spg info
        try:
            spg_sym_opt, spg_num_opt = opt_struc.get_space_group_info(
                symprec=rin.symprec)
        except TypeError:
            spg_num_opt = 0
            spg_sym_opt = None
        # ------ out opt_struc
        out_poscar({cid:opt_struc}, './data/opt_POSCARS')
        try:
            out_cif(opt_struc, cid, work_path,
                    './data/opt_CIFS.cif', rin.symprec)
        except TypeError:
            logger.warning('failed to write opt_CIF')
    # ---------- error
    else:
        spg_num_opt = 0
        spg_sym_opt = None

    # ---------- register opt_struc
    opt_struc_data[cid] = opt_struc
    pkl_data.save_opt_struc(opt_struc_data)
    # ---------- return
    #return spg_sym, spg_num, spg_sym_opt, spg_num_opt

    # ---------- register rslt
    if rin.algo in ['RS', 'LAQA']:
        rslt_data.loc[cid] = [spg_num, spg_sym, spg_num_opt, spg_sym_opt,
                              energy, magmom, check_opt]
    elif rin.algo == 'BO':
        rslt_data.loc[cid] = [n_selection,
                              spg_num, spg_sym, spg_num_opt, spg_sym_opt,
                              energy, magmom, check_opt]
    elif rin.algo in ['EA']:
        rslt_data.loc[cid] = [gen,
                              spg_num, spg_sym, spg_num_opt, spg_sym_opt,
                              energy, magmom, check_opt]
    elif rin.algo in ['EA-vc']:
        rslt_data.loc[cid] = [gen,
                              spg_num, spg_sym, spg_num_opt, spg_sym_opt,
                              energy, ef, magmom, check_opt]
    pkl_data.save_rslt(rslt_data)
    if rin.algo != 'EA-vc':
        out_rslt(rslt_data)
    else:    # EA-vc
        out_rslt(rslt_data, order_ef=True)

    # ---------- return
    return opt_struc_data, rslt_data


def update_status(cid, id_queueing, id_running, operation):

    # ---------- read stat
    stat = io_stat.stat_read()

    # ---------- update status
    if operation == 'submit':
        id_running.append(cid)
        id_queueing.remove(cid)
        io_stat.set_stage(stat, cid, 1)
    elif operation == 'fin':
        if cid in id_queueing:
            id_queueing.remove(cid)
        if cid in id_running:
            id_running.remove(cid)
        io_stat.clean_id(stat, cid)
    else:
        logger.error('operation is wrong')
        raise SystemExit(1)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save id_queueing and id_running
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)

    # ---------- return
    return id_queueing, id_running


def mv_fin(cid):
    if not os.path.isdir(f'work/fin/{cid}'):
        shutil.move(f'work/{cid}', 'work/fin/')
    else:    # rename for recalc
        for i in itertools.count(1):
            if not os.path.isdir(f'work/fin/{cid}_{i}'):
                shutil.move(f'work/{cid}',
                            f'work/fin/{cid}_{i}')
                break


def next_gen_EA(
        rin,
        id_queueing,
        id_running,
        gen,
        go_next_sg,
        init_struc_data,
        opt_struc_data,
        rslt_data,
        ea_vc_data=None,
        struc_mol_id=None,
    ):

    # ---------- log
    logger.info(f'Done generation {gen}')

    # ---------- flag for next selection or generation
    if not go_next_sg:
        logger.info('\nEA is ready')
        os.remove('lock_cryspy')
        raise SystemExit()

    # ---------- check point 3
    if rin.stop_chkpt == 3:
        logger.info('\nStop at check point 3: EA is ready')
        os.remove('lock_cryspy')
        raise SystemExit()

    # ---------- maxgen_ea
    if 0 < rin.maxgen_ea <= gen:
        if rin.algo == 'EA-vc':
            # ------ when gen reaches maxgen_ea,  next generation is not created
            #        so convex hull is calculated here
            #        update only convex hull and hdist, not elite_struc and elite_fitness
            c_rslt = rslt_data[rslt_data['Gen'] == gen]
            c_ids = c_rslt.index.values    # current IDs [array]
            ef_all = rslt_data['Ef_eV_atom'].to_dict()    # formation energy of all structures
            from ..EA.calc_hull import calc_convex_hull_2d
            nat_data, ratio_data, hdist_data = ea_vc_data
            hdist = calc_convex_hull_2d(rin, ratio_data, ef_all, c_ids, gen)
            out_hdist(gen, hdist, ratio_data)
            hdist_data[gen] = hdist
            pkl_data.save_hdist_data(hdist_data)
        logger.info(f'\nReached maxgen_ea: {rin.maxgen_ea}')
        os.remove('lock_cryspy')
        raise SystemExit()

    # ---------- EA
    backup_cryspy()
    from ..EA import ea_next_gen
    ea_next_gen.next_gen(
        rin,
        id_queueing,
        id_running,
        gen,
        init_struc_data,
        opt_struc_data,
        rslt_data,
        ea_vc_data,
        struc_mol_id,
    )


'''
Structure file for OpenMX
written by H. Sawahata 2020/03/09
info at hikaruri.jp
'''

from pymatgen.core import Structure
from pymatgen.core.units import Length


import re


def extract_cell_parameters_from_outfile(filename):
    # ---------- last Cell vectors a1, a2, a3
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines_cell = None
    for i,line in enumerate(lines):
        if re.search('a1', line):
            lines_cell = [line[7:37],lines[i+1][7:37],
                lines[i+2][7:37]]
            break
    return lines_cell


def extract_cell_parameters_from_infile(filename):
    # ---------- last Cell vectors a1, a2, a3
    # case of not optimized Cell
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines_cell = None
    for i,line in enumerate(lines):
        if re.search('<Atoms.UnitVectors', line):
            lines_cell = [lines[i+1],lines[i+2],lines[i+3]]
            break
    return lines_cell

def extract_atomic_positions_from_outfile(filename, nat):
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    # ---------- last atomic positions
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines_atom = None
    for i, line in enumerate(lines):
        if re.search('final', line):
            ibegin = i + 4
            iend = ibegin + natot
            lines_atom = lines[ibegin:iend]
            break
    return lines_atom

def extract_atomic_positions_from_infile(rin, filename):
    # ---------- last atomic positions
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines_atom = []
    for i,line in enumerate(lines):
        if re.search('<Atoms.SpeciesAndCoordinates', line):
            for j in range(0,rin.natot):
                Atom_Row = lines[i+j+1].split()
                Row = ' '.join(Atom_Row[0:5])
                lines_atom.append(Row)
            break
    return lines_atom

def from_lines(lines_cell, lines_atom):
    # ---------- lattice
    lattice = [[float(x) for x in line.split()]
        for line in lines_cell[0:3]]
    # ---------- species & coordinates
    species = []
    coords = []
    for line in lines_atom:
        fields = line.split()
        species.append(fields[1])
        coords.append([float(x) for x in fields[2:5]])
    structure = Structure(lattice, species, coords)
    return structure

def write(rin, structure, output, mode='w'):
    # ---------- get in POSCAR format
    poscar = structure.to(fmt='poscar')
    lines = poscar.split('\n')[:-1]
    lines = [line+'\n' for line in lines]
    # ---------- write in OMX format (yet)
    with open(output, mode) as f:
        f.write('<Atoms.UnitVectors\n')
        for line in lines[2:5]:
            f.write(line)
        f.write('Atoms.UnitVectors>\n')
        f.write('<Atoms.SpeciesAndCoordinates\n')  
        for i,line in enumerate(lines[8:]):
            fields = line.split()
            up     = str(rin.upSpin[fields[3]])
            down   = str(rin.downSpin[fields[3]])
            f.write(str(i+1)+' '+fields[3]+'  '+' '.join(fields[0:3])+' '+up+' '+down+'\n')
        f.write('Atoms.SpeciesAndCoordinates>\n')  



from logging import getLogger

import numpy as np

from ...util.struc_util import check_distance, sort_by_atype


logger = getLogger('cryspy')


def gen_substitution(
        atype,
        mindist,
        struc_data,
        sp,
        n_subs,
        nat_data,
        ll_nat,
        ul_nat,
        id_start=None,
        symprec=0.01,
        maxcnt_ea=50,
        target='random',
    ):
    '''

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    struc_data (dict): {id: structure data}
    sp (instance): instance of SelectParents class
    n_subs (int): number of structures to generate by substitution
    nat_data (dict): {id: nat}
    ll_nat (tuple): lower limit of nat, e.g. (1, 1)
    ul_nat (tuple): upper limit of nat, e.g. (8, 8)
    id_start (int): start ID for new structures
    symprec (float): tolerance for symmetry
    maxcnt_ea (int): maximum number of trial in substitution
    target (str): only 'random' for now

    # ---------- return
    children (dict): {id: structure data}
    parents (dict): {id: (id of parent_A, )}
    operation (dict): {id: 'strain'}
    '''

    # ---------- initialize
    struc_cnt = 0
    children = {}
    #children_mol_id = {}
    parents = {}
    operation = {}

    # ---------- id_offset
    if id_start is None:
        cid = max(struc_data.keys()) + 1
    else:
        if id_start < (max(struc_data.keys()) + 1):
            logger.error('id_start is already included in structure ID of the data')
        else:
            cid = id_start

    # ---------- generate structures by substitution
    while struc_cnt < n_subs:
        # ------ select parents
        pid_A, = sp.get_parents(n_parent=1)    # comma for list[0]
        parent_A = struc_data[pid_A]
        # ------ check nat limit
        atype_avail_elim = []
        atype_avail_add = []
        for i, at in enumerate(atype):
            if nat_data[pid_A][i] > ll_nat[i]:
                atype_avail_elim.append(at)
            if nat_data[pid_A][i] < ul_nat[i]:
                atype_avail_add.append(at)
        if len(atype_avail_elim) == 1:
            # if atype_avail_elim is ['Na'], 'Na' should be removed from atype_avail_add
            at = atype_avail_elim[0]
            if at in atype_avail_add:
                atype_avail_add.remove(at)
        if len(atype_avail_add) == 1:
            # if atype_avail_add is ['Na'], 'Na' should be removed from atype_avail_elim
            at = atype_avail_add[0]
            if at in atype_avail_elim:
                atype_avail_elim.remove(at)
        if len(atype_avail_add) == 0:
            logger.warning('Substitution: reached nat limit (ul_nat).')
            logger.warning('Change parent')
            continue
        if len(atype_avail_elim) == 0:
            logger.warning('Substitution: reached nat limit (ll_nat).')
            logger.warning('Change parent')
            continue
        child = gen_child(atype, mindist, parent_A, atype_avail_add, atype_avail_elim,
                          maxcnt_ea, target)
        # ------ success
        if child is not None:
            children[cid] = child
            parents[cid] = (pid_A, )
            operation[cid] = 'substitution'
            try:
                spg_sym, spg_num = child.get_space_group_info(symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            logger.info(f'Structure ID {cid:>6} was generated'
                f' from {pid_A:>6} by substitution.'
                f' Space group: {spg_num:>3} {spg_sym}')
            cid += 1
            struc_cnt += 1

    # ---------- return
    return children, parents, operation


def gen_child(atype, mindist, parent_A, atype_avail_add, atype_avail_elim,
              maxcnt_ea=50, target='random'):
    '''
        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    parent_A (Structure): pymatgen Structure object
    atype_avail_add (list): available atom type for addition
    atype_avail_elim (list): available atom type for elimination
    maxcnt_ea (int): maximum number of trial in crossover
    target (str): only 'random' for now

    # ---------- return
    (if success) child (Structure): pymatgen Structure object
    (if fail) None
    '''

    # ---------- initialize
    cnt = 0

    # ---------- generate structure by substitution
    while True:
        child = parent_A.copy()    # keep original structure
        if target == 'random':
            # ------ random choice for atom type
            at = np.random.choice(atype_avail_elim)
            if at in atype_avail_add:
                atype_avail_add.remove(at)
            if len(atype_avail_add) == 0:
                logger.warning('Substitution: no atype_avail_add, retry.')
                cnt += 1
                if cnt >= maxcnt_ea:
                    logger.warning('Substitution: cnt >= maxcnt_ea.')
                    logger.warning('Change parent')
                    return None    # change parent
                continue
            else:
                # ------ elimination
                aindx = [i for i, site in
                            enumerate(child) if site.species_string == at]
                elim_indx = np.random.choice(aindx, 1)  # ", 1" to get array
                coords = child[elim_indx[0]].frac_coords    # coords of eliminated atom
                child.remove_sites(elim_indx)
                # ------ addition
                at = np.random.choice(atype_avail_add)
                child.append(species=at, coords=coords)
        # ---------- check mindist
        success, mindist_ij, dist = check_distance(child, atype, mindist)
        if success:
            child = sort_by_atype(child, atype)
            return child
        else:
            type0 = atype[mindist_ij[0]]
            type1 = atype[mindist_ij[1]]
            logger.warning(f'mindist in addition: {type0} - {type1}, {dist}. retry.')
            cnt += 1
            if cnt >= maxcnt_ea:
                logger.warning('Substitution: cnt >= maxcnt_ea.')
                logger.warning('Change parent')
                return None    # change parent



from logging import getLogger

from ..calc_dscrpt.FP.calc_FP import calc_fp


logger = getLogger('cryspy')


def select_descriptor(rin, struc_data):
    # ---------- fingerprint
    if rin.dscrpt == 'FP':
        logger.info('Calculate descriptors: FingerPrint by Valle and Oganov')
        # ------ calc fingerprint
        descriptors = calc_fp(struc_data, rin.atype, rin.fp_rmax,
                              rin.fp_npoints, rin.fp_sigma)
        return descriptors
    else:
        logger.error('Now FP only')
        raise SystemExit(1)



'''
Change input variables in cryspy.in
'''

import configparser


def read_config():
    config = configparser.ConfigParser()
    config.read('cryspy.in')
    return config


def write_config(config):
    with open('cryspy.in', 'w') as f:
        config.write(f)


def change_input(config, section, key, value):
    config.set(section, key, f'{value}')



from logging import getLogger

from ..IO import io_stat, pkl_data


logger = getLogger('cryspy')

def restart(rin, prev_nstruc):
    # ---------- load laqa data
    id_queueing = pkl_data.load_id_queueing()
    laqa_step = pkl_data.load_laqa_step()
    laqa_struc = pkl_data.load_laqa_struc()
    laqa_energy = pkl_data.load_laqa_energy()
    laqa_bias = pkl_data.load_laqa_bias()
    laqa_score = pkl_data.load_laqa_score()

    # ---------- initialize for appended ID
    for i in range(prev_nstruc, rin.tot_struc):
        laqa_step[i] = []
        laqa_struc[i] = []
        laqa_energy[i] = []
        laqa_bias[i] = []
        laqa_score[i] = [float('inf')]
        id_queueing.append(i)
    logger.info('Append scores and id_queueing')

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save for LAQA
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_laqa_step(laqa_step)
    pkl_data.save_laqa_struc(laqa_struc)
    pkl_data.save_laqa_energy(laqa_energy)
    pkl_data.save_laqa_bias(laqa_bias)
    pkl_data.save_laqa_score(laqa_score)



from logging import getLogger
import os
import shutil

from pymatgen.core import Structure
from pymatgen.io.vasp import Kpoints

from ...IO.out_results import out_kpts
from ...IO import pkl_data


logger = getLogger('cryspy')


def next_stage_vasp(rin, stage, work_path, kpt_data, cid):
    # ---------- skip_flag
    skip_flag = False

    # ---------- rename VASP files at the current stage
    vasp_files = ['POSCAR', 'KPOINTS', 'CONTCAR',
                  'OUTCAR', 'OSZICAR', 'vasprun.xml']
    for file in vasp_files:
        if not os.path.isfile(work_path + file):
            logger.error('Not found ' + work_path + file)
            raise SystemExit(1)
        os.rename(work_path + file, work_path + f'stage{stage}_' + file)

    # ---------- cp CONTCAR POSCAR
    shutil.copyfile(work_path + f'stage{stage}_CONTCAR',
                    work_path + 'POSCAR')

    # ---------- remove STOPCAR
    if os.path.isfile(work_path+'STOPCAR'):
        os.remove(work_path+'STOPCAR')

    # ---------- KPOINTS for the next stage using pymatgen
    try:
        structure = Structure.from_file(work_path+'POSCAR')
    except ValueError:
        skip_flag = True
        kpt_data[cid].append(['skip'])
        pkl_data.save_kpt(kpt_data)
        out_kpts(kpt_data)
        logger.info(f'    error in VASP,  skip structure {cid}')
        return skip_flag, kpt_data
    # kppvol[0]: <--> stage 1, kppvol[1] <--> stage2, ...
    #   so (stage - 1): current stage, stage: next stage in kppvol
    kpoints = Kpoints.automatic_density_by_vol(structure=structure,
                                               kppvol=rin.kppvol[stage],
                                               force_gamma=rin.force_gamma)
    kpoints.write_file(work_path+'KPOINTS')

    # ---------- kpt_data
    kpt_data[cid].append(kpoints.kpts[0])
    pkl_data.save_kpt(kpt_data)
    out_kpts(kpt_data)

    # ---------- cp INCAR_? from ./calc_in for the next stage: (stage + 1)
    fincar = f'./calc_in/INCAR_{stage + 1}'
    shutil.copyfile(fincar, work_path+'INCAR')

    # ---------- return
    return skip_flag, kpt_data


def next_struc_vasp(rin, structure, cid, work_path, kpt_data):
    # ---------- copy files
    calc_inputs = ['POTCAR', 'INCAR']
    for f in calc_inputs:
        ff = f+'_1' if f == 'INCAR' else f
        if not os.path.isfile('./calc_in/' + ff):
            logger.error('Could not find ./calc_in/' + ff)
            raise SystemExit(1)
        # ------ e.g. cp ./calc_in/INCAR_1 work/1/INCAR
        shutil.copyfile('./calc_in/'+ff, work_path+f)

    # ---------- generate POSCAR
    structure.to(fmt='poscar', filename=work_path+'POSCAR')
    if not os.path.isfile(work_path+'POSCAR'):
        logger.error(f'Could not find {work_path}POSCAR')
        raise SystemExit(1)

    # ---------- Change the title of POSCAR
    with open(work_path+'POSCAR', 'r') as f:
        lines = f.readlines()
    lines[0] = f'ID_{cid}\n'
    with open(work_path+'POSCAR', 'w') as f:
        for line in lines:
            f.write(line)

    # ---------- generate KPOINTS using pymatgen
    kpoints = Kpoints.automatic_density_by_vol(structure=structure,
                                               kppvol=rin.kppvol[0],
                                               force_gamma=rin.force_gamma)
    kpoints.write_file(work_path+'KPOINTS')

    # ---------- kpt_data
    kpt_data[cid] = []    # initialize
    kpt_data[cid].append(kpoints.kpts[0])
    pkl_data.save_kpt(kpt_data)
    out_kpts(kpt_data)

    # ---------- return
    return kpt_data



'''
Restart Bayesian optimization
'''

from .select_descriptor import select_descriptor
from ..IO import pkl_data


def restart(rin, init_struc_data, prev_nstruc):
    # ---------- load BO data
    init_dscrpt_data = pkl_data.load_init_dscrpt_data()

    # ---------- get additional struc data
    tmp_struc_data = {cid: init_struc_data[cid] for cid in range(
        prev_nstruc, len(init_struc_data))}

    # ---------- calc descriptor
    tmp_dscrpt = select_descriptor(rin, tmp_struc_data)

    # ---------- update
    init_dscrpt_data.update(tmp_dscrpt)

    # ---------- save
    pkl_data.save_init_dscrpt_data(init_dscrpt_data)



from logging import getLogger

from .gen_struc_EA.select_parents import SelectParents
from .gen_struc_EA.crossover import gen_crossover
from .gen_struc_EA.permutation import gen_permutation
from .gen_struc_EA.strain import gen_strain
from .gen_struc_EA.addition import gen_addition
from .gen_struc_EA.elimination import gen_elimination
from .gen_struc_EA.substitution import gen_substitution
#from .gen_struc_EA.rotation import gen_rotation
from ..IO import pkl_data
from ..IO.out_results import out_nat_data
from ..RS.rs_gen import gen_random
from ..util.utility import check_fwpath
from ..util.struc_util import set_mindist, get_mol_data, out_poscar, get_nat

# ---------- import later
#from ..RS.gen_struc_RS import gen_pyxtal
#from ..RS.gen_struc_RS import random_generation


logger = getLogger('cryspy')


def child_gen(rin, ranking, fittest, struc_data, init_struc_data,
              struc_mol_id=None, ea_vc_data=None):

    # ---------- instantiate SelectParents class
    sp = SelectParents(ranking)    # after set_xxx, we can use sp.get_parents(n_parent)
    if rin.slct_func == 'TNM':
        sp.set_tournament(rin.t_size)
    else:
        sp.set_roulette(fittest, rin.a_rlt, rin.b_rlt, rin.fit_reverse)
        logger.debug(f'cumulative fitness in roulette: {sp.cum_fit}')

    # ---------- set mindist
    logger.info('# -- mindist')
    mindist = set_mindist(rin.atype, rin.mindist, rin.mindist_factor, rin.struc_mode)
    if rin.struc_mode == 'mol_bs':
        mindist_dummy = set_mindist(rin.atype,
                                    rin.mindist_mol_bs,
                                    rin.mindist_mol_bs_factor,
                                    rin.struc_mode,
                                    dummy=True,
                                    mol_file=rin.mol_file,
                                    mpi_rank=0)
    # ---------- initialize
    children = {}
    #children_mol_id = {}
    parents = {}
    operation = {}
    id_start = rin.tot_struc
    vc = True if rin.algo == 'EA-vc' else False

    # ---------- Crossover
    if rin.n_crsov > 0:
        if rin.struc_mode not in ['mol', 'mol_bs']:
            co_children, co_parents, co_operation = gen_crossover(
                rin.atype,
                rin.nat,
                mindist,
                struc_data,
                sp,
                rin.n_crsov,
                id_start,
                rin.symprec,
                rin.crs_lat,
                rin.nat_diff_tole,
                rin.maxcnt_ea,
                vc,
                rin.ll_nat,
                rin.ul_nat,
                struc_mol_id=None,
                molecular=False,
            )
        else:
            logger.error('Crossover is not implemented for mol or mol_bs')
            # co = Crossover(rin, mindist)
            # eagen.gen_crossover(rin, co, struc_mol_id, molecular=True)
        # ------ update
        children.update(co_children)
        parents.update(co_parents)
        operation.update(co_operation)
        id_start += rin.n_crsov

    # ---------- Permutation
    if rin.n_perm > 0:
        if rin.struc_mode not in ['mol', 'mol_bs']:
            pm_children, pm_parents, pm_operation = gen_permutation(
                rin.atype,
                mindist,
                struc_data,
                sp,
                rin.n_perm,
                id_start,
                rin.symprec,
                rin.ntimes,
                rin.maxcnt_ea,
                struc_mol_id=None,
                molecular=False,
            )
        else:
            logger.error('Permutation is not implemented for mol or mol_bs')
            # pm = Permutation(mindist)
            # eagen.gen_permutation(rin, pm, struc_mol_id, molecular=True)
        # ------ update
        children.update(pm_children)
        parents.update(pm_parents)
        operation.update(pm_operation)
        id_start += rin.n_perm

    # ---------- Strain
    if rin.n_strain > 0:
        if rin.struc_mode not in ['mol', 'mol_bs']:
            st_children, st_parents, st_operation = gen_strain(
                rin.atype,
                mindist,
                struc_data,
                sp,
                rin.n_strain,
                id_start,
                rin.symprec,
                rin.sigma_st,
                rin.maxcnt_ea,
                struc_mol_id=None,
                molecular=False,
                protect_mol_struc=True,
            )
        else:
            logger.error('Strain is not implemented for mol or mol_bs')
            # st = Strain(mindist)
            # eagen.gen_strain(rin, st, struc_mol_id, protect_mol_struc=True)
        # ------ update
        children.update(st_children)
        parents.update(st_parents)
        operation.update(st_operation)
        id_start += rin.n_strain

    # ---------- EA-vc
    if rin.algo == 'EA-vc':
        nat_data, ratio_data, _ = ea_vc_data

        # ------ Addition
        if rin.n_add > 0:
            if rin.struc_mode not in ['mol', 'mol_bs']:
                ad_children, ad_parents, ad_operation = gen_addition(
                    rin.atype,
                    mindist,
                    struc_data,
                    sp,
                    rin.n_add,
                    nat_data,
                    rin.ul_nat,
                    id_start,
                    rin.symprec,
                    rin.maxcnt_ea,
                    rin.target,
                )
            else:
                logger.error('Addition is not implemented for mol or mol_bs')
                raise SystemExit(1)
            # ------ update
            children.update(ad_children)
            parents.update(ad_parents)
            operation.update(ad_operation)
            id_start += rin.n_add

        # ------ Elimination
        if rin.n_elim > 0:
            if rin.struc_mode not in ['mol', 'mol_bs']:
                el_children, el_parents, el_operation = gen_elimination(
                    rin.atype,
                    struc_data,
                    sp,
                    rin.n_elim,
                    nat_data,
                    rin.ll_nat,
                    id_start,
                    rin.symprec,
                    rin.target,
                )
            else:
                logger.error('Elimination is not implemented for mol or mol_bs')
                raise SystemExit(1)
            # ------ update
            children.update(el_children)
            parents.update(el_parents)
            operation.update(el_operation)
            id_start += rin.n_elim

        # ------ Substitution
        if rin.n_subs > 0:
            if rin.struc_mode not in ['mol', 'mol_bs']:
                sb_children, sb_parents, sb_operation = gen_substitution(
                    rin.atype,
                    mindist,
                    struc_data,
                    sp,
                    rin.n_subs,
                    nat_data,
                    rin.ll_nat,
                    rin.ul_nat,
                    id_start,
                    rin.symprec,
                    rin.maxcnt_ea,
                    rin.target,
                )
            else:
                logger.error('Substitution is not implemented for mol or mol_bs')
                raise SystemExit(1)
            # ------ update
            children.update(sb_children)
            parents.update(sb_parents)
            operation.update(sb_operation)
            id_start += rin.n_subs

    # ---------- Rotation
    # if rin.struc_mode in ['mol', 'mol_bs']:
    #     if rin.n_rotation > 0:
    #         rot = Rotation(mindist)
    #         eagen.gen_rotation(rin, struc_mol_id, rot=rot)

    # ---------- write init_POSCARS
    out_poscar(children, './data/init_POSCARS')

    # ---------- update init_struc_data
    init_struc_data.update(children)
    # if rin.struc_mode in ['mol', 'mol_bs']:
    #     struc_mol_id.update(children_mol_id)

    # ---------- save EA-vc_data.pkl
    if rin.algo == 'EA-vc':
        # ea_vc_data is already loaded above
        for cid, struc in children.items():
            nat_data[cid], ratio_data[cid] = get_nat(struc, rin.atype)
        pkl_data.save_nat_data(nat_data)
        pkl_data.save_ratio_data(ratio_data)

    # ---------- random generation
    if rin.n_rand > 0:
        tmp_struc_data, tmp_mol_id = gen_random(
            rin,
            rin.n_rand,
            id_start,
            comm=None,
            mpi_rank=0,
            mpi_size=1,
        )
        # ------ update
        init_struc_data.update(tmp_struc_data)
        # if rin.struc_mode in ['mol', 'mol_bs']:
        #     struc_mol_id.update(tmp_mol_id)
        # ------ save EA-vc_data.pkl
        if rin.algo == 'EA-vc':
            # ea_vc_data is already loaded above
            for cid, struc in tmp_struc_data.items():
                nat_data[cid], ratio_data[cid] = get_nat(struc, rin.atype)
            pkl_data.save_nat_data(nat_data)
            pkl_data.save_ratio_data(ratio_data)
        # ------ write init_POSCARS
        out_poscar(tmp_struc_data, './data/init_POSCARS')

    # ---------- save init_struc_data
    pkl_data.save_init_struc(init_struc_data)
    # if rin.struc_mode in ['mol', 'mol_bs']:
    #     pkl_data.save_struc_mol_id(struc_mol_id)

    # ---------- out nat_data
    if rin.algo == 'EA-vc':
        out_nat_data(nat_data, rin.atype)

    # ----------return
    return init_struc_data, parents, operation
    #return init_struc_data, parents, operation, struc_mol_id


# not used in this version
# this is used in adj_comp.py
# def check_vcnat(rin, child):
#     from ..util.struc_util import get_nat
#     nat, ratio = get_nat(child, rin.atype)

#     if len(rin.atype) == 2:
#         for i in range(len(rin.atype)):
#             if (nat[i] >= rin.ll_nat[i]) and (nat[i]<= rin.ul_nat[i]):
#                 check_nat = True
#             else:
#                 #check_nat = False
#                 return False

#         return True
        
#     elif len(rin.atype) == 3:
#         SystemExit(1) #temporary
#     else:
#         SystemExit(1) #temporary


from logging import getLogger


logger = getLogger('cryspy')


def out_input(rin):
    logger.info('[basic]')
    for key in rin.__annotations__.keys():
        if key == 'struc_mode':
            logger.info('')
            logger.info('[structure]')
        if key == 'stop_chkpt':
            logger.info('')
            logger.info('[option]')
        if rin.algo == 'BO' and key == 'nselect_bo':
            logger.info('')
            logger.info('[BO]')
        if rin.algo == 'LAQA' and key == 'nselect_laqa':
            logger.info('')
            logger.info('[LAQA]')
        if rin.algo in ['EA', 'EA-vc'] and key == 'n_pop':
            logger.info('')
            logger.info('[EA]')
        if rin.calc_code == 'VASP' and key == 'kpt_flag':
            logger.info('')
            logger.info('[VASP]')
        if rin.calc_code == 'QE' and key == 'kpt_flag':
            logger.info('')
            logger.info('[QE]')
        if rin.calc_code == 'OMX' and key == 'kpt_flag':
            logger.info('')
            logger.info('[OMX]')
        if rin.calc_code == 'soiap' and key == 'kpt_flag':
            logger.info('')
            logger.info('[soiap]')
        if rin.calc_code == 'LAMMPS' and key == 'kpt_flag':
            logger.info('')
            logger.info('[LAMMPS]')
        if rin.calc_code == 'ASE' and key == 'kpt_flag':
            logger.info('')
            logger.info('[ASE]')
        if rin.calc_code == 'ext' and key == 'kpt_flag':
            logger.info('')
            logger.info('[ext]')
        if getattr(rin, key) is not None:
            logger.info(f'{key} = {getattr(rin, key)}')


'''
Utility for CrySPY
'''

from datetime import datetime
from logging import getLogger, StreamHandler, FileHandler, Formatter, DEBUG, INFO, WARNING
import os
import shutil
import subprocess
import sys


logger = getLogger('cryspy')

def get_version():
    return '1.3.0'


def set_logger(noprint=False, debug=False, logfile=None, errfile=None, debugfile=None):
    # ---------- level and formatter
    logger.setLevel(DEBUG)
    fmt = Formatter("[%(asctime)s][%(module)s][%(levelname)s] %(message)s")

    # ---------- stream handler for log
    if not noprint:
        shandler = StreamHandler(stream=sys.stdout)
        shandler.setFormatter(fmt)
        if debug:
            shandler.setLevel(DEBUG)
        else:
            shandler.setLevel(INFO)
        logger.addHandler(shandler)

    # ---------- file handler for log (level == INFO)
    if logfile is not None:
        fhandler = FileHandler(logfile)
        fhandler.setFormatter(fmt)
        fhandler.addFilter(lambda record: record.levelno == INFO)
        logger.addHandler(fhandler)

    # ---------- file handler for error (WARNING <= level)
    if errfile is not None:
        ehandler = FileHandler(errfile)
        ehandler.setFormatter(fmt)
        ehandler.setLevel(WARNING)
        logger.addHandler(ehandler)

    # ---------- file handler for debug (level == DEBUG)
    if debug:
        if debugfile is not None:
            dhandler = FileHandler(debugfile)
            dhandler.setFormatter(fmt)
            dhandler.addFilter(lambda record: record.levelno == DEBUG)
            logger.addHandler(dhandler)


def check_fwpath(fwpath):
    if fwpath is None:
        # ---------- check if find_wy is in your path
        sr = subprocess.run(['which', 'find_wy'], capture_output=True, text=True)
        fwpath = sr.stdout.strip()    # to delete \n
        if fwpath == '':
            logger.error('There is no find_wy program in your path')
            raise SystemExit(1)
    else:
        # ---------- check fwpath written in cryspy.in
        if not os.path.isfile(fwpath):
            logger.error(f'There is no find_wy program in {fwpath}')
            raise SystemExit(1)
    return fwpath


def backup_cryspy():
    # ---------- make directory
    dname = datetime.now().strftime("%Y%m%d_%H%M%S")
    dst = 'backup/' + dname + '/'
    os.makedirs(dst, exist_ok=True)

    # ---------- file/directory list
    flist = ['cryspy.in', 'cryspy.stat', 'err_cryspy', 'log_cryspy']
    dlist = ['calc_in', 'data', 'ext']

    # ---------- backup
    for f in flist:
        if os.path.isfile(f):
            shutil.copy2(f, dst)
    for d in dlist:
        if os.path.isdir(d):
            shutil.copytree(d, dst + d)

    # ---------- print
    logger.info('Backup data')


def clean_cryspy(skip_yes=False):
    # ---------- yes/no
    if not skip_yes:
        while True:
            choice = input("Are you sure you want to clean the data? 'yes' or 'no' [y/n]: ").lower()
            if choice in ['y', 'yes']:
                break
            elif choice in ['n', 'no']:
                return

    # ---------- file/directory list
    fdlist = ['cryspy.stat', 'debug_cryspy', 'err_cryspy', 'log_cryspy', 'lock_cryspy',
              'data', 'work', 'ext', 'tmp_calc_FP', 'tmp_gen_struc']

    # ---------- clean
    dname = datetime.now().strftime("%Y%m%d_%H%M%S")
    dst = 'trash/' + dname + '/'
    os.makedirs(dst, exist_ok=True)
    for fd in fdlist:
        if os.path.exists(fd):
            shutil.move(fd, dst)



from logging import getLogger

import numpy as np

#from .adj_comp import operation_atoms, convex_hull_check


logger = getLogger('cryspy')


def gen_elimination(
        atype,
        struc_data,
        sp,
        n_elim,
        nat_data,
        ll_nat,
        id_start=None,
        symprec=0.01,
        target='random',
    ):
    '''

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    struc_data (dict): {id: structure data}
    sp (instance): instance of SelectParents class
    n_elim (int): number of structures to generate by elimination
    nat_data (dict): {id: nat}
    ll_nat (tuple): lower limit of nat, e.g. (1, 1)
    id_start (int): start ID for new structures
    symprec (float): tolerance for symmetry
    target (str): only 'random' for now

    # ---------- return
    children (dict): {id: structure data}
    parents (dict): {id: (id of parent_A, )}
    operation (dict): {id: 'strain'}
    '''

    # ---------- initialize
    struc_cnt = 0
    children = {}
    #children_mol_id = {}
    parents = {}
    operation = {}

    # ---------- id_offset
    if id_start is None:
        cid = max(struc_data.keys()) + 1
    else:
        if id_start < (max(struc_data.keys()) + 1):
            logger.error('id_start is already included in structure ID of the data')
        else:
            cid = id_start

    # ---------- generate structures by elimination
    while struc_cnt < n_elim:
        # ------ select parents
        pid_A, = sp.get_parents(n_parent=1)    # comma for list[0]
        parent_A = struc_data[pid_A]
        # ------ check nat limit
        atype_avail = []
        for i, at in enumerate(atype):
            if nat_data[pid_A][i] > ll_nat[i]:
                atype_avail.append(at)
        if len(atype_avail) == 0:
            logger.warning('Elimination: reached nat limit (ll_nat). cannot add atoms')
            logger.warning('Change parent')
            continue
        child = gen_child(parent_A, atype_avail, target)
        # ------ success
        if child is not None:
            children[cid] = child
            parents[cid] = (pid_A, )
            operation[cid] = 'elimination'
            try:
                spg_sym, spg_num = child.get_space_group_info(symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            logger.info(f'Structure ID {cid:>6} was generated'
                f' from {pid_A:>6} by elimination.'
                f' Space group: {spg_num:>3} {spg_sym}')
            cid += 1
            struc_cnt += 1

    # ---------- return
    return children, parents, operation


def gen_child(parent_A, atype_avail, target='random'):
    '''
    parent_A (Structure): pymatgen Structure object
    atype_avail (list): available atom type for elimination
    target (str): only 'random' for now

    # ---------- return
    child (Structure): pymatgen Structure object
    '''

    # ---------- keep original structure
    child = parent_A.copy()    # keep original structure
    # ---------- elimination
    if target == 'random':
        # ------ random choice for atom type
        at = np.random.choice(atype_avail)
        # ------ random choice for atom index
        aindx = [i for i, site in enumerate(child) if site.species_string == at]
        elim_indx = np.random.choice(aindx, 1)  # ", 1" to get array
        # ------ remove atom
        child.remove_sites(elim_indx)
    # ---------- not implemented yet
    # elif tgt in ['depop', 'overpop']:
    #     section = convex_hull_check(rin)
    #     child = operation_atoms(rin, 'elimination', child, section)
    # ---------- return
    #            no need to check distance in elimination
    return child



'''
Control jobs in soiap
'''

from logging import getLogger
import os
import shutil

from . import structure as soiap_structure


logger = getLogger('cryspy')

def next_stage_soiap(rin, stage, work_path, nat):
    # ---------- skip_flag
    skip_flag = False

    # ---------- rename soiap files at the current stage
    soiap_files = [rin.soiap_infile, rin.soiap_outfile, rin.soiap_cif,
                   'log.struc', 'log.tote', 'log.frc', 'log.strs']
    for file in soiap_files:
        if not os.path.isfile(work_path + file):
            logger.error('Not found ' + work_path + file)
            raise SystemExit(1)
        os.rename(work_path + file, work_path + f'stage{stage}_' + file)

    # ---------- copy the input file from ./calc_in for the next stage
    finfile = './calc_in/' + rin.soiap_infile + f'_{stage + 1}'
    shutil.copyfile(finfile, work_path+rin.soiap_infile)

    # ---------- generate the CIF file
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        with open(work_path+f'stage{stage}_log.struc', 'r') as f:
            lines = f.readlines()
            lines = lines[-(natot+5):]
        structure = soiap_structure.from_file(lines, nat)
    except ValueError:
        skip_flag = True
        logger.warning('    error in soiap,  skip this structure')
        return skip_flag
    with open(work_path + f'stage{stage}_' + rin.soiap_cif, 'r') as f:
        lines = f.readlines()
    title = lines[0][5:]    # string following 'data_'
    soiap_structure.write(structure,
                          work_path+rin.soiap_cif,
                          symprec=rin.symprec,
                          title=title)

    # ---------- return
    return skip_flag


def next_struc_soiap(rin, structure, cid, work_path):
    # ---------- copy files
    calc_inputs = [rin.soiap_infile]
    for f in calc_inputs:
        ff = f+'_1' if f == rin.soiap_infile else f
        if not os.path.isfile('./calc_in/'+ff):
            logger.error('Could not find ./calc_in/'+ff)
            raise SystemExit(1)
        shutil.copyfile('./calc_in/'+ff, work_path+f)

    # ---------- generate the CIF file
    soiap_structure.write(structure,
                          work_path+rin.soiap_cif,
                          symprec=rin.symprec,
                          title=f'ID_{cid}')



from logging import getLogger

import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import ConvexHull


logger = getLogger('cryspy')


def calc_convex_hull_2d(rin, ratio_data, ef_all, c_ids, gen):
    '''
    Input:
        ratio_data [dict]: ratio of all structures, {ID: [ratio list], ...}
        ef_all [dict]: formation energy of all structures, {ID: Ef, ...}
        c_ids [array]: ID array of current generation structures
        gen [int]: current generation

    Return:
        hdist [dict]: hull distance of all structures, {ID: distance, ...}
    '''
    # ---------- initialize
    ratio_chull =  [0.0, 1.0]    # only ef < 0 for calculation of convex hull
    ef_chull =  [0.0, 0.0]       # only ef < 0 for calculation of convex hull

    # ---------- dict to list
    for cid in ratio_data:
        if ef_all[cid] < 0.0:
            ratio_chull.append(ratio_data[cid][0])
            ef_chull.append(ef_all[cid])

    # ---------- no negative Ef
    #            in this case, hull distance is equivalent to Ef
    if len(ef_chull) == 2:    # only end points
        # np.nan --> np.inf in hdist
        hdist = {cid: np.inf if np.isnan(ef_all[cid]) else ef_all[cid] for cid in ef_all}
        draw_convex_hull_2d(rin, None, ratio_data, ef_all, c_ids, gen)
        return hdist

    # ---------- calc convex hull
    points = list(zip(ratio_chull, ef_chull))
    hull = ConvexHull(points)
    vpoints = hull.points[hull.vertices]    # hull.vertices: index of vertices in hull.points
    vpoints = np.vstack((vpoints, vpoints[0]))    # just for plot. vpoints[0] should be [1, 0] by ConvexHull()

    # ---------- hull distance
    hdist = {
        cid: np.inf if np.isnan(ef_all[cid])
        else hull_distance_2d(ratio_data[cid][0], ef_all[cid], hull.equations)
        for cid in ef_all
    }

    # ---------- draw convex hull
    draw_convex_hull_2d(rin, vpoints, ratio_data, ef_all, c_ids, gen)

    # ---------- return
    return hdist


def hull_distance_2d(x, y, equations):
    '''
    equations: [eq0, eq1, eq2, ...], eq0: [a, b, c] for a*x + b*y + c = 0

    Find the distance from all equations and adopt the minimum distance.
    '''
    hdists = []
    for eq in equations:
        if eq[0] != 0.0:
            dist = y - (-eq[0]*x - eq[2])/eq[1]
            if dist < 0.0:
                logger.warning('hdist <= 0.0, check hull distance.')
            hdists.append(dist)

    return min(hdists)


def draw_convex_hull_2d(rin, vpoints, ratio_data, ef_all, c_ids, gen):
    # ---------- setting
    plt.rcParams.update(_set_params())

    # ---------- fig
    fig, ax = plt.subplots(1, 1)

    # ---------- hline
    ax.axhline(y=0, xmin=0, xmax=1, color='black', linestyle='--')

    # ---------- label
    ax.set_xlabel('$x$ in '+f'{rin.atype[0]}'+'$_{x}$'+f'{rin.atype[1]}'+'$_{1-x}$')
    ax.set_ylabel('Formation energy (eV/atom)')

    # ---------- lim
    min_ef = min(ef_all.values())
    ymin = min_ef - 0.01 if min_ef < 0 else -0.01
    ax.set_ylim(ymin, 0.05)

    # ---------- plot ef_all
    for cid in ef_all:
        if np.isnan(ef_all[cid]):
            continue
        if cid in c_ids:
            ax.plot(ratio_data[cid][0], ef_all[cid], 'o', ms=12, mew=2.0, c='C2', alpha=0.8)
        else:
            ax.plot(ratio_data[cid][0], ef_all[cid], 'o', ms=12, mew=2.0, c='C0', alpha=0.8)

    # ---------- plot convex hull
    if vpoints is None:    # no negative Ef, plot only end points
        ax.plot([0, 1], [0, 0], 'o', ms=12, mew=2.0, c='C1')
    else:
        ax.plot(vpoints[1:, 0], vpoints[1:, 1], '-o', ms=12, mew=2.0, c='C1')

    # ---------- save figure
    fig.savefig(f'./data/convex_hull/conv_hull_gen_{gen}.png', bbox_inches='tight')


def _set_params():
    rcParams_dict = {
        # ---------- figure
        'figure.figsize': [8, 6],
        'figure.dpi': 120,
        'figure.facecolor': 'white',
        # ---------- axes
        'axes.grid': True,
        'axes.linewidth': 1.5,
        # ---------- ticks
        'xtick.direction': 'in',
        'ytick.direction': 'in',
        'xtick.major.width': 1.0,
        'ytick.major.width': 1.0,
        'xtick.major.size': 8.0,
        'ytick.major.size': 8.0,
        # ---------- lines
        'lines.linewidth': 2.5,
        'lines.markersize': 12,
        # ---------- grid
        'grid.linestyle': ':',
        # ---------- font
        'font.family': 'Times New Roman',
        'mathtext.fontset': 'cm',
        #'mathtext.fontset': 'stix',
        'font.size': 20,
        'axes.labelsize': 26,
        'legend.fontsize': 26,
        'svg.fonttype': 'path',  # Embed characters as paths
        #'svg.fonttype': 'none',  # Assume fonts are installed on the machine
        'pdf.fonttype': 42,  # embed fonts in PDF using type42 (True type)
    }
    return rcParams_dict



from logging import getLogger
import os

import pandas as pd

from ..IO.out_results import out_ea_info, out_nat_data
from ..IO import io_stat, pkl_data
from ..util.struc_util import get_nat


logger = getLogger('cryspy')

def initialize(rin, init_struc_data, rslt_data):
    # ---------- log
    logger.info('# ---------- Initialize evolutionary algorithm')
    logger.info('# ------ Generation 1')
    logger.info(f'{rin.tot_struc} structures by random')

    # ---------- initialize
    gen = 1
    id_queueing = [i for i in range(rin.tot_struc)]
    id_running = []
    # ------ ea_info
    if rin.algo == 'EA':
        ea_info = pd.DataFrame(columns=[
                                    'Gen',
                                    'Population',
                                    'Crossover',
                                    'Permutation',
                                    'Strain',
                                    'Random',
                                    'Elite',
                                    'crs_lat',
                                    'slct_func',
                                ])
        ea_info.iloc[:, 0:7] = ea_info.iloc[:, 0:7].astype(int)
        tmp_info = pd.DataFrame([[
                                    1,
                                    rin.tot_struc,
                                    0,
                                    0,
                                    0,
                                    rin.tot_struc,
                                    0,
                                    rin.crs_lat,
                                    rin.slct_func,
                                ]], columns=ea_info.columns)
        ea_info = pd.concat([ea_info, tmp_info], axis=0, ignore_index=True)
    if rin.algo == 'EA-vc':
        ea_info = pd.DataFrame(columns=[
                                    'Gen',
                                    'Population',
                                    'Crossover',
                                    'Permutation',
                                    'Strain',
                                    'Addition',
                                    'Elimination',
                                    'Substitution',
                                    'Random',
                                    'Elite',
                                    'crs_lat',
                                    'slct_func',
                                ])
        ea_info.iloc[:, 0:10] = ea_info.iloc[:, 0:10].astype(int)
        tmp_info = pd.DataFrame([[
                            1,
                            rin.tot_struc,
                            0,
                            0,
                            0,
                            0,
                            0,
                            0,
                            rin.tot_struc,
                            0,
                            rin.crs_lat,
                            rin.slct_func,
                    ]], columns=ea_info.columns)
        ea_info = pd.concat([ea_info, tmp_info], axis=0, ignore_index=True)
    out_ea_info(ea_info)
    # ------ ea_origin
    ea_origin = pd.DataFrame(columns=['Gen', 'Struc_ID', 'Operation', 'Parent'])
    ea_origin.iloc[:, 0:2] = ea_origin.iloc[:, 0:2].astype(int)
    for cid in range(rin.tot_struc):
        tmp_origin = pd.DataFrame([[1, cid, 'random', None]],
                                  columns=ea_origin.columns)
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ elite
    elite_struc = None
    elite_fitness = None
    # ------ rslt_data
    rslt_data['Gen'] = pd.Series(dtype=int)
    rslt_data = rslt_data[[
                    'Gen',
                    'Spg_num',
                    'Spg_sym',
                    'Spg_num_opt',
                    'Spg_sym_opt',
                    'E_eV_atom',
                    'Magmom',
                    'Opt',
                ]]
    if rin.algo == 'EA-vc':
        rslt_data['Ef_eV_atom'] = pd.Series(dtype='float64')
        rslt_data = rslt_data[[
                        'Gen',
                        'Spg_num',
                        'Spg_sym',
                        'Spg_num_opt',
                        'Spg_sym_opt',
                        'E_eV_atom',
                        'Ef_eV_atom',
                        'Magmom',
                        'Opt',
                    ]]
    # ------ nat, ratio, etc for EA-vc
    if rin.algo == 'EA-vc':
        nat_data = {}      # {ID: [nat], ...}
        ratio_data = {}    # {ID: [ratio], ...}
        hdist_data = {}    # {gen: {ID: hdist, ...}, ...}
        for cid, struc in init_struc_data.items():
            tmp_nat, tmp_ratio = get_nat(struc, rin.atype)
            nat_data[cid] = tmp_nat
            ratio_data[cid] = tmp_ratio
        out_nat_data(nat_data, rin.atype)
        os.makedirs('data/convex_hull', exist_ok=True)

    # ---------- save
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)
    pkl_data.save_gen(gen)
    pkl_data.save_elite_struc(elite_struc)
    pkl_data.save_elite_fitness(elite_fitness)
    pkl_data.save_ea_info(ea_info)
    pkl_data.save_ea_origin(ea_origin)
    pkl_data.save_rslt(rslt_data)
    if rin.algo == 'EA-vc':
        pkl_data.save_nat_data(nat_data)
        pkl_data.save_ratio_data(ratio_data)
        pkl_data.save_hdist_data(hdist_data)

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_common(stat, 'generation', gen)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)



'''
Calculate fingerprint by Valle and Oganov using DScribe
'''

from dscribe.descriptors import ValleOganov
from pymatgen.io.ase import AseAtomsAdaptor


def calc_fp(struc_data, atype, fp_rmax, fp_npoints, fp_sigma):
    '''
    # ---------- args
    struc_data (dict): {ID: pymatgen Structure, ...}
    atype (tuple): e.g. ('Na', 'Cl')
    fp_rmax (float): cutoff radius
    fp_npoints (int): number of points
    fp_sigma (float): standard deviation in Gaussian smearing
    '''

    # ---------- initialize
    descriptors = {}

    # ---------- calc fingerprint
    for cid, struc in struc_data.items():
        atoms = AseAtomsAdaptor.get_atoms(struc)
        descriptors[cid] = calc_fp_describe(atoms, atype, fp_rmax, fp_npoints, fp_sigma)

    # ---------- return
    return descriptors



def calc_fp_describe(atoms, atype, r_cut, n, sigma, function='distance'):
    vo = ValleOganov(
        species=atype,
        function=function,
        sigma=sigma,
        n=n,
        r_cut=r_cut,
    )
    vo_fp = vo.create(atoms)    # np.array
    return vo_fp



import configparser
from logging import getLogger
import random

import pandas as pd

from .select_descriptor import select_descriptor
from ..IO import io_stat, pkl_data


logger = getLogger('cryspy')


def initialize(rin, init_struc_data, rslt_data):
    # ---------- log
    logger.info('# ---------- Selection: 1')

    # ---------- check init_struc_data
    if None in init_struc_data.values():
        raise ValueError('init_struc_data includes None')

    # ---------- initialize
    n_selection = 1
    id_running = []
    id_select_hist = []
    bo_mean = {}
    bo_var = {}
    bo_score = {}

    # ---------- rslt_data, add and sort
    rslt_data['Select'] = pd.Series(dtype=int)
    rslt_data = rslt_data[[
                    'Select',
                    'Spg_num',
                    'Spg_sym',
                    'Spg_num_opt',
                    'Spg_sym_opt',
                    'E_eV_atom',
                    'Magmom',
                    'Opt',
                ]]
    pkl_data.save_rslt(rslt_data)

    # ---------- random select
    all_id = [i for i in range(len(init_struc_data))]
    if rin.manual_select_bo:
        # ------ manual select bo
        x = ' '.join(str(i) for i in rin.manual_select_bo)
        logger.info(f'Manual select: {x}')
        nselect = rin.nselect_bo - len(rin.manual_select_bo)
        id_queueing = list(rin.manual_select_bo[:])    # shallow copy
        if 0 < nselect:
            diff_id = list(set(all_id) - set(id_queueing))
            id_queueing.extend(random.sample(diff_id, nselect))
        # ------ delete the value for manual_select_bo in cryspy.in
        config = configparser.ConfigParser()
        config.read('cryspy.in')
        config.set('BO', 'manual_select_bo', '')
        with open('cryspy.in', 'w') as f:
            config.write(f)
    else:
        id_queueing = random.sample(all_id, rin.nselect_bo)

    # ---------- id_select_hist
    id_select_hist.append(id_queueing[:])    # append shallow copy

    # ---------- calc descriptor
    init_dscrpt_data = select_descriptor(rin, init_struc_data)
    opt_dscrpt_data = {}  # initialize in dict

    # ---------- save for BO
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)
    pkl_data.save_n_selection(n_selection)
    pkl_data.save_id_select_hist(id_select_hist)
    pkl_data.save_init_dscrpt_data(init_dscrpt_data)
    pkl_data.save_opt_dscrpt_data(opt_dscrpt_data)
    pkl_data.save_bo_mean(bo_mean)
    pkl_data.save_bo_var(bo_var)
    pkl_data.save_bo_score(bo_score)

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_common(stat, 'selection', n_selection)
    io_stat.set_id(stat, 'selected_id', id_queueing)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- out and log
    x = ' '.join(str(a) for a in id_queueing)
    logger.info(f'selected_id: {x}')




'''
Collect results in VASP
'''

from logging import getLogger
import os
import xml.etree.ElementTree as ET

import numpy as np
from pymatgen.core import Structure

from ...util import constants
from ...IO import pkl_data


logger = getLogger('cryspy')

def collect_vasp(cid, work_path, nat):
    # ---------- check optimization
    check_opt = check_opt_vasp(work_path+'OUTCAR')
    # ---------- obtain energy and magmom
    energy, magmom = get_energy_magmom_vasp(work_path, nat)
    if np.isnan(energy):
        logger.warning(f'    Structure ID {cid},'
              ' could not obtain energy from OSZICAR')
    # ---------- collect CONTCAR
    try:
        opt_struc = Structure.from_file(work_path+'CONTCAR')
    except Exception:
        opt_struc = None
    # ---------- check
    if np.isnan(energy):
        opt_struc = None
    if opt_struc is None:
        energy = np.nan
        magmom = np.nan
    # ---------- remove STOPCAR
    if os.path.isfile(work_path+'STOPCAR'):
        os.remove(work_path+'STOPCAR')
    # ---------- return
    return opt_struc, energy, magmom, check_opt


def check_opt_vasp(file_path):
    try:
        with open(file_path, 'r') as f:
            lines = f.readlines()
        check_opt = 'not_yet'
        for line in lines:
            if 'reached required accuracy' in line:
                check_opt = 'done'
    except Exception:
        check_opt = 'no_file'
    return check_opt


def get_energy_magmom_vasp(work_path, nat):
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    # ---------- obtain energy and magmom
    energy = np.nan
    magmom = np.nan
    try:
        with open(work_path+'OSZICAR', 'r') as foszi:
            oszi = foszi.readlines()
        if 'F=' in oszi[-1]:
            energy = float(oszi[-1].split()[2])    # free energy (eV/cell)
            energy = energy/float(natot)       # eV/atom
            if 'mag=' in oszi[-1]:
                magmom = float(oszi[-1].split()[-1])    # total magnetic moment
    except Exception:
        pass
    # ---------- return
    return energy, magmom


def get_energy_step_vasp(energy_step_data, cid, work_path, nat):
    '''
    get energy step data in eV/atom

    energy_step_data[ID][stage][step]
    energy_step_data[ID][0] <-- stage 1
    energy_step_data[ID][1] <-- stage 2
    '''
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    # ---------- get energy step from vasprun
    try:
        # ------ read file
        tree = ET.parse(work_path+'vasprun.xml')
        root = tree.getroot()
        # ------ children nodes: calculation
        cals = root.findall('calculation')
        # ------ init.
        energy_step = []
        # ------ loop for relaxation step
        for cal in cals:
            eng = cal.find('energy')    # first 'energy' child node
            fr_eng = eng.find('i')    # first 'i' tag is free energy

            if fr_eng.attrib['name'] == 'e_fr_energy':
                energy_step.append(fr_eng.text)
            else:
                logger.error('bug in get_energy_step_vasp')
                raise SystemExit(1)
        # ------ list, str --> array
        energy_step = np.array(energy_step, dtype='float')/float(natot)
    except Exception as e:
        energy_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse in energy_step')

    # ---------- append energy_step
    if energy_step_data.get(cid) is None:
        energy_step_data[cid] = []    # initialize
    energy_step_data[cid].append(energy_step)

    # ---------- save energy_step_data
    pkl_data.save_energy_step(energy_step_data)

    # ---------- return
    return energy_step_data


def get_struc_step_vasp(struc_step_data, cid, work_path):
    '''
    get structure step data

    # ---------- args
    struc_step_data: (dict) the key is structure ID

    struc_step_data[ID][stage][step]
    struc_step_data[ID][0] <-- stage 1
    struc_step_data[ID][1] <-- stage 2
    '''
    # ---------- get struc step from vasprun
    try:
        # ------ read file
        tree = ET.parse(work_path+'vasprun.xml')
        root = tree.getroot()
        # ------ get atom list
        atoms = root.findall("atominfo/array[@name='atoms']/set/rc")
        atomlist = []
        for atom in atoms:
            atomlist.append(atom.find('c').text)
        # ------ children nodes: calculation
        cals = root.findall('calculation')
        # ------ init.
        struc_step = []
        # ------ loop for relaxation step
        for cal in cals:
            # -- lattice
            basis = cal.findall("structure/crystal/varray[@name='basis']/v")
            lattice = []
            for a in basis:
                lattice.append([float(x) for x in a.text.split()])
            # -- positions
            positions = cal.findall("structure/varray/[@name='positions']/v")
            incoord = []
            for a in positions:
                incoord.append([float(x) for x in a.text.split()])
            # -- structure in pymatgen format
            struc = Structure(lattice, atomlist, incoord)
            # -- append
            struc_step.append(struc)
    except Exception as e:
        struc_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse in struc_step')

    # ---------- append struc_step
    if struc_step_data.get(cid) is None:
        struc_step_data[cid] = []    # initialize
    struc_step_data[cid].append(struc_step)

    # ---------- save struc_step_data
    pkl_data.save_struc_step(struc_step_data)

    # ---------- return
    return struc_step_data


def get_force_step_vasp(force_step_data, cid, work_path):
    '''
    get force step data in eV/angstrom

    # ---------- args
    force_step_data: (dict) the key is structure ID

    force_step_data[ID][stage][step]
    force_step_data[ID][0] <-- stage 1
    force_step_data[ID][1] <-- stage 2
    '''
    # ---------- get force step from vasprun
    try:
        # ------ read file
        tree = ET.parse(work_path+'vasprun.xml')
        root = tree.getroot()
        # ------ children nodes: calculation
        cals = root.findall('calculation')
        # ------ init.
        force_step = []
        # ------ loop for ralaxation step
        for cal in cals:
            varrays = cal.findall('varray')
            # -- init
            force = []
            # -- varrays[0]: force, varrays[1]: stress
            for varray in varrays:
                vs = varray.findall('v')
                # loop for v
                for v in vs:
                    if varray.attrib['name'] == 'forces':
                        force.append(v.text.split())
            # -- list, str --> array
            force = np.array(force, dtype='float')
            # -- appned force_step
            force_step.append(force)
    except Exception as e:
        force_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse in force_step')

    # ---------- append force_step
    if force_step_data.get(cid) is None:
        force_step_data[cid] = []    # initialize
    force_step_data[cid].append(force_step)

    # ---------- save force_step_data
    pkl_data.save_force_step(force_step_data)

    # ---------- return
    return force_step_data


def get_stress_step_vasp(stress_step_data, cid, work_path):
    '''
    get stress step data in eV/ang**3

    # ---------- args
    stress_step_data: (dict) the key is structure ID

    stress_step_data[ID][stage][step]
    stress_step_data[ID][0] <-- stage 1
    stress_step_data[ID][1] <-- stage 2
    '''
    # ---------- get stress step from vasprun
    try:
        # ------ read file
        tree = ET.parse(work_path+'vasprun.xml')
        root = tree.getroot()
        # ------ children nodes: calculation
        cals = root.findall('calculation')
        # ------ init.
        stress_step = []
        # ------ loop for ralaxation step
        for cal in cals:
            varrays = cal.findall('varray')
            # -- init
            stress = []
            # -- varrays[0]: force, varrays[1]: stress
            for varray in varrays:
                vs = varray.findall('v')
                # loop for v
                for v in vs:
                    if varray.attrib['name'] == 'stress':
                        stress.append(v.text.split())
            # -- list, str --> array
            stress = np.array(stress, dtype='float')
            # -- kbar --> eV/ang**3
            stress = stress * constants.KBAR2eV_ANG3
            # -- appned stress_step
            stress_step.append(stress)
    except Exception as e:
        stress_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse in stress_step')

    # ---------- append stress_step
    if stress_step_data.get(cid) is None:
        stress_step_data[cid] = []    # initialize
    stress_step_data[cid].append(stress_step)

    # ---------- save stress_step_data
    pkl_data.save_stress_step(stress_step_data)

    # ---------- return
    return stress_step_data



from ..IO import io_stat, pkl_data


def restart(rin, prev_nstruc):
    # ---------- load rs data
    id_queueing = pkl_data.load_id_queueing()

    # ---------- append id_queueing
    id_queueing.extend([i for i in range(prev_nstruc, rin.tot_struc)])

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save
    pkl_data.save_id_queueing(id_queueing)

    # ---------- ext
    if rin.calc_code == 'ext':
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('out\n')




# def display_rslt():
#     path = './data/cryspy_rslt_energy_asc'
#     with open(path) as f:
#         rslt_read = f.read()
#     print(rslt_read)
import pickle


def display_rslt(cid):
    path = './data/pkl_data/rslt_data.pkl'
    with open(path, 'rb') as f:
        rslt_data = pickle.load(f)

    sum = len(rslt_data)

    if cid == 'all':
        print(rslt_data)
    elif cid <= sum and cid >= 0:
        print(rslt_data.loc[cid])
    else:
        print("This number structure does not exist.")



from logging import getLogger
import os

from ..IO import io_stat, pkl_data
from ..IO.out_results import out_laqa_id_hist


logger = getLogger('cryspy')


def next_selection(rin, laqa_id_data, laqa_data):
    # ---------- laqa_id_data and laqa_data
    (id_queueing, id_running, id_select_hist) = laqa_id_data
    (tot_step_select, laqa_step, laqa_struc,
     laqa_energy, laqa_bias, laqa_score) = laqa_data

    # ---------- LAQA selection
    for k, v in sorted(laqa_score.items(), key=lambda x: -x[1][-1]):
        if v[-1] == -float('inf'):
            break
        else:
            id_queueing.append(k)
            if len(id_queueing) == rin.nselect_laqa:
                break

    # ---------- done LAQA
    if len(id_queueing) == 0:
        logger.info('\nDone all structures!')
        os.remove('lock_cryspy')
        raise SystemExit()

    # ---------- append id_select_hist and out
    id_select_hist.append(id_queueing[:])    # append shallow copy
    out_laqa_id_hist(id_select_hist)

    # ---------- tot_step_select for next selection
    tot_step_select.append(0)

    # ---------- save
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)
    pkl_data.save_id_select_hist(id_select_hist)
    pkl_data.save_tot_step_select(tot_step_select)
    pkl_data.save_laqa_step(laqa_step)
    pkl_data.save_laqa_struc(laqa_struc)
    pkl_data.save_laqa_energy(laqa_energy)
    pkl_data.save_laqa_bias(laqa_bias)
    pkl_data.save_laqa_score(laqa_score)

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_common(stat, 'selection', len(id_select_hist))
    io_stat.set_id(stat, 'selected_id', id_queueing)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- out and log
    logger.info(f'# ---------- Selection {len(id_select_hist)}')
    if len(id_queueing) > 30:
        logger.info(f'selected_id: {len(id_queueing)} IDs')
    else:
        x = ' '.join(str(a) for a in id_queueing)
        logger.info(f'selected_id: {x}')
